import jax
import jax.numpy as jnp
from functools import partial
from collections import namedtuple
import warnings
from typing import Literal, Union, overload, Any

import numpy as np

import mbirjax as mj
from mbirjax import TomographyModel, tomography_utils

ConeBeamParamNames = mj.ParamNames | Literal['angles', 'source_detector_dist', 'source_iso_dist', 'recon_slice_offset']


class ConeBeamModel(TomographyModel):
    """
    A class designed for handling forward and backward projections in a cone beam geometry, extending the
    :ref:`TomographyModelDocs`. This class offers specialized methods and parameters tailored for cone beam setups.

    This class inherits all methods and properties from the :ref:`TomographyModelDocs` and may override some
    to suit parallel beam geometrical requirements. See the documentation of the parent class for standard methods
    like setting parameters and performing projections and reconstructions.

    Parameters not included in the constructor can be set using the set_params method of :ref:`TomographyModelDocs`.
    Refer to :ref:`TomographyModelDocs` documentation for a detailed list of possible parameters.

    Args:
        sinogram_shape (tuple):
            Shape of the sinogram as a tuple in the form `(views, rows, channels)`, where 'views' is the number of
            different projection angles, 'rows' correspond to the number of detector rows, and 'channels' index columns of
            the detector that are assumed to be aligned with the rotation axis.
        angles (ndarray or jax array):
            A 1D array of projection angles, in radians, specifying the angle of each projection relative to the origin.
        source_detector_dist (float): Distance between the X-ray source and the detector in units of ALU.
        source_iso_dist (float): Distance between the X-ray source and the center of rotation in units of ALU.

    Note:
        One additional parameter for ConeBeamModel that can be set using set_params() is

        **recon_slice_offset** (float, default=0) -
        Vertical offset of the image in ALU. If recon_slice_offset is positive, we reconstruct the region below iso.

    See Also
    --------
    TomographyModel : The base class from which this class inherits.
    """

    DIRECT_RECON_VIEW_BATCH_SIZE = TomographyModel.DIRECT_RECON_VIEW_BATCH_SIZE

    def __init__(self, sinogram_shape, angles, source_detector_dist, source_iso_dist):

        self.bp_psf_radius = 1
        self.entries_per_cylinder_batch = 128
        self.slice_range_length = 0
        angles = jnp.asarray(angles)
        view_params_name = 'angles'

        super().__init__(sinogram_shape, angles=angles, source_detector_dist=source_detector_dist,
                         source_iso_dist=source_iso_dist, view_params_name=view_params_name, recon_slice_offset=0.0)

    @overload
    def get_params(self, parameter_names: Union[ConeBeamParamNames, list[ConeBeamParamNames]]) -> Any: ...

    def get_params(self, parameter_names) -> Any:
        return super().get_params(parameter_names)

    def get_magnification(self):
        """
        Returns the magnification for the cone beam geometry.

        Returns:
            magnification = source_detector_dist / source_iso_dist
        """
        source_detector_dist, source_iso_dist = self.get_params(['source_detector_dist', 'source_iso_dist'])
        if jnp.isinf(source_detector_dist):
            magnification = 1
        else:
            magnification = source_detector_dist / source_iso_dist
        return magnification

    def verify_valid_params(self):
        """
        Check that all parameters are compatible for a reconstruction.

        Note:
            Raises ValueError for invalid parameters.
        """
        super().verify_valid_params()
        sinogram_shape, angles = self.get_params(['sinogram_shape', 'angles'])

        if angles.shape[0] != sinogram_shape[0]:
            error_message = "Number view dependent parameter vectors must equal the number of views. \n"
            error_message += "Got {} for length of view-dependent parameters and "
            error_message += "{} for number of views.".format(angles.shape[0], sinogram_shape[0])
            raise ValueError(error_message)

        # Check for cone angle > 45 degrees
        source_detector_dist, delta_det_row, det_row_offset = \
            self.get_params(['source_detector_dist', 'delta_det_row', 'det_row_offset'])
        half_detector_height = delta_det_row * sinogram_shape[1] / 2 + jnp.abs(det_row_offset)
        if half_detector_height > source_detector_dist:
            warnings.warn('Cone angle is more than 45 degrees.  This will likely produce recon artifacts.')

        # TODO:  Check for recon volume extending into the source
        # # Check for a potential division by zero or very small denominator
        # if (source_to_iso_dist - y) < 1e-3>:
        #     raise ValueError("Invalid geometry: Recon volume extends too close to source.")

    def get_geometry_parameters(self):
        """
        Function to get a list of the primary geometry parameters for cone beam projection.

        Returns:
            namedtuple of required geometry parameters.
        """
        # First get the parameters managed by ParameterHandler
        geometry_param_names = \
            ['delta_det_row', 'delta_det_channel', 'det_row_offset', 'det_channel_offset',
             'source_detector_dist', 'delta_voxel', 'recon_slice_offset']
        geometry_param_values = self.get_params(geometry_param_names)

        # Then get additional parameters:
        geometry_param_names += ['magnification', 'psf_radius', 'bp_psf_radius',
                                 'entries_per_cylinder_batch', 'slice_range_length']
        geometry_param_values.append(self.get_magnification())
        geometry_param_values.append(self.get_psf_radius())
        geometry_param_values.append(self.bp_psf_radius)
        geometry_param_values.append(self.entries_per_cylinder_batch)
        geometry_param_values.append(self.slice_range_length)

        # Then create a namedtuple to access parameters by name in a way that can be jit-compiled.  
        GeometryParams = namedtuple('GeometryParams', geometry_param_names)
        geometry_params = GeometryParams(*tuple(geometry_param_values))

        return geometry_params

    def get_psf_radius(self):
        """
        Compute the integer radius of the PSF kernel for cone beam projection.
        """
        delta_det_row, delta_det_channel, source_detector_dist, recon_shape, delta_voxel = self.get_params(
            ['delta_det_row', 'delta_det_channel', 'source_detector_dist', 'recon_shape', 'delta_voxel'])
        magnification = self.get_magnification()

        # Compute minimum detector pitch
        delta_det = jnp.minimum(delta_det_row, delta_det_channel)

        # Compute maximum magnification
        if jnp.isinf(source_detector_dist):
            max_magnification = 1
            min_magnification = 1
        else:
            source_to_iso_dist = source_detector_dist / magnification
            # This isn't exactly the closest pixel since we're not accounting for rotation but for realistic cases it shouldn't matter.
            source_to_closest_pixel = source_to_iso_dist - 0.5 * jnp.maximum(recon_shape[0], recon_shape[1])*delta_voxel
            max_magnification = source_detector_dist / source_to_closest_pixel
            source_to_farthest_pixel = source_to_iso_dist + 0.5 * jnp.maximum(recon_shape[0], recon_shape[1])*delta_voxel
            min_magnification = source_detector_dist / source_to_farthest_pixel

        # Compute the maximum number of detector rows/channels on either side of the center detector hit by a voxel
        psf_radius = int(jnp.ceil(jnp.ceil((delta_voxel * max_magnification / delta_det)) / 2))
        # Then repeat for the back projection from detector elements to voxels.
        # The voxels closest to the detector will be covered the most by a given detector element.
        # With magnification=1, the number of voxels per element would be delta_det / delta_voxel
        max_voxels_per_detector = delta_det / (min_magnification * delta_voxel)
        self.bp_psf_radius = int(jnp.ceil(jnp.ceil(max_voxels_per_detector) / 2))

        self.slice_range_length = int(1 + 2 * self.bp_psf_radius + \
                                  jnp.ceil(self.entries_per_cylinder_batch * max_voxels_per_detector))

        return psf_radius

    def auto_set_recon_shape(self, sinogram_shape, no_compile=True, no_warning=False):
        """ Compute the automatic recon shape cone beam reconstruction.
        """
        delta_det_row, delta_det_channel = self.get_params(['delta_det_row', 'delta_det_channel'])
        delta_voxel = self.get_params('delta_voxel')
        num_det_rows, num_det_channels = sinogram_shape[1:3]
        magnification = self.get_magnification()

        num_recon_rows = int(jnp.round(num_det_channels * ((delta_det_channel / delta_voxel) / magnification)))
        num_recon_cols = num_recon_rows
        num_recon_slices = int(jnp.round(num_det_rows * ((delta_det_row / delta_voxel) / magnification)))

        recon_shape = (num_recon_rows, num_recon_cols, num_recon_slices)
        self.set_params(no_compile=no_compile, no_warning=no_warning, recon_shape=recon_shape)

    @staticmethod
    @partial(jax.jit, static_argnames='projector_params')
    def forward_project_pixel_batch_to_one_view(voxel_values, pixel_indices, angle, projector_params):
        """
        Forward project a set of voxels determined by indices into the flattened array of size num_rows x num_cols.

        Args:
            voxel_values (jax array):  2D array of shape (num_indices, num_slices) of voxel values, where
                voxel_values[i, j] is the value of the voxel in slice j at the location determined by indices[i].
            pixel_indices (jax array of int):  1D vector of indices into flattened array of size num_rows x num_cols.
            angle (float):  Angle for this view
            projector_params (namedtuple):  tuple of (sinogram_shape, recon_shape, get_geometry_params())

        Returns:
            jax array of shape (num_det_rows, num_det_channels)
        """
        recon_shape = projector_params.recon_shape
        num_recon_slices = recon_shape[2]
        if voxel_values.shape[0] != pixel_indices.shape[0] or len(voxel_values.shape) < 2 or \
                voxel_values.shape[1] != num_recon_slices:
            raise ValueError('voxel_values must have shape[0:2] = (num_indices, num_slices)')

        vertical_fan_projector = ConeBeamModel.forward_vertical_fan_pixel_batch_to_one_view
        horizontal_fan_projector = ConeBeamModel.forward_horizontal_fan_pixel_batch_to_one_view

        new_voxel_values = vertical_fan_projector(voxel_values, pixel_indices, angle, projector_params)
        sinogram_view = horizontal_fan_projector(new_voxel_values, pixel_indices, angle, projector_params)

        return sinogram_view

    @staticmethod
    def forward_vertical_fan_pixel_batch_to_one_view(voxel_values, pixel_indices, angle, projector_params):
        """
        Apply a fan beam forward projection in the vertical direction separately to each voxel determined by indices
        into the flattened array of size num_rows x num_cols.  This returns an array corresponding to the same pixel
        locations, but using values obtained from the projection of the original voxel cylinders onto a detector column,
        so the output array has size (len(pixel_indices), num_det_rows).

        Args:
            voxel_values (jax array):  2D array of shape (num_pixels, num_recon_slices) of voxel values, where
                voxel_values[i, j] is the value of the voxel in slice j at the location determined by indices[i].
            pixel_indices (jax array of int):  1D vector of shape (len(pixel_indices), ) holding the indices into
                the flattened array of size num_rows x num_cols.
            angle (float):  Angle for this view
            projector_params (namedtuple):  tuple of (sinogram_shape, recon_shape, get_geometry_params())

        Returns:
            jax array of shape (num_pixels, num_det_rows)
        """
        pixel_map = jax.vmap(ConeBeamModel.forward_vertical_fan_one_pixel_to_one_view,
                             in_axes=(0, 0, None, None))
        new_pixels = pixel_map(voxel_values, pixel_indices, angle, projector_params)

        return new_pixels

    @staticmethod
    def forward_horizontal_fan_pixel_batch_to_one_view(voxel_values, pixel_indices, angle, projector_params):
        """
        Apply a horizontal fan beam transformation to a set of voxel cylinders. These cylinders are assumed to have
        slices aligned with detector rows, so that a horizontal fan beam maps a cylinder slice to a detector row.
        This function returns the resulting sinogram view.

        Args:
            voxel_values (jax array):  2D array of shape (num_pixels, num_recon_slices) of voxel values, where
                voxel_values[i, j] is the value of the voxel in slice j at the location determined by indices[i].
            pixel_indices (jax array of int):  1D vector of shape (len(pixel_indices), ) holding the indices into
                the flattened array of size num_rows x num_cols.
            angle (float):  Angle for this view
            projector_params (namedtuple):  tuple of (sinogram_shape, recon_shape, get_geometry_params())

        Returns:
            jax array of shape (num_det_rows, num_det_channels)
        """

        # Get all the geometry parameters - we use gp since geometry parameters is a named tuple and we'll access
        # elements using, for example, gp.delta_det_channel, so a longer name would be clumsy.
        gp = projector_params.geometry_params
        num_views, num_det_rows, num_det_channels = projector_params.sinogram_shape

        # Get the data needed for horizontal projection
        n_p, n_p_center, W_p_c, cos_alpha_p_xy = ConeBeamModel.compute_horizontal_data(pixel_indices, angle, projector_params)
        L_max = jnp.minimum(1, W_p_c)

        # Allocate the sinogram array
        sinogram_view = jnp.zeros((num_det_rows, num_det_channels))

        # Do the horizontal projection
        for n_offset in jnp.arange(start=-gp.psf_radius, stop=gp.psf_radius + 1):
            n = n_p_center + n_offset
            abs_delta_p_c_n = jnp.abs(n_p - n)
            L_p_c_n = jnp.clip((W_p_c + 1) / 2 - abs_delta_p_c_n, 0, L_max)
            A_chan_n = gp.delta_voxel * L_p_c_n / cos_alpha_p_xy
            A_chan_n *= (n >= 0) * (n < num_det_channels)
            sinogram_view = sinogram_view.at[:, n].add(A_chan_n.reshape((1, -1)) * voxel_values.T)

        return sinogram_view

    @staticmethod
    def forward_vertical_fan_one_pixel_to_one_view(voxel_cylinder, pixel_index, angle, projector_params):
        """
        Apply a fan beam forward projection in the vertical direction to the pixel determined by indices
        into the flattened array of size num_rows x num_cols.  This returns a vector obtained from the projection of
        the original voxel cylinder onto a detector column, so the output vector has length num_det_rows.

        Args:
            voxel_cylinder (jax array):  1D array of shape (num_recon_slices, ) of voxel values, where
                voxel_cylinder[j] is the value of the voxel in slice j at the location determined by pixel_index.
            pixel_index (int):  Index into the flattened array of size num_rows x num_cols.
            angle (float):  Angle for this view.
            projector_params (namedtuple):  tuple of (sinogram_shape, recon_shape, get_geometry_params())

        Returns:
            jax array of shape (num_det_rows,)

        Note:
            This is a helper function used in vmap in :meth:`ConeBeamModel.forward_vertical_fan_pixel_batch_to_one_view`
        This method has the same signature and output as that method, except single int pixel_index is used
        in place of the 1D pixel_indices, and likewise only a single voxel cylinder is returned.
        """
        # Get all the geometry parameters - we use gp since geometry parameters is a named tuple and we'll access
        # elements using, for example, gp.delta_det_channel, so a longer name would be clumsy.
        gp = projector_params.geometry_params
        num_views, num_det_rows, num_det_channels = projector_params.sinogram_shape
        recon_shape = projector_params.recon_shape
        num_slices = voxel_cylinder.shape[0]

        # From pixel index, compute y and pixel_mag
        y, pixel_mag = ConeBeamModel.compute_y_mag_for_pixel(pixel_index, angle, recon_shape, projector_params)

        # The code above depends only on the pixel - a single point.  z is a potentially large vector
        # Here we compute cos_phi_p:  1 / cos_phi_p determines the projection length through a voxel
        # For computational efficiency, we use that to scale the voxel_cylinder values.
        # TODO:  possibly convert to a jitted function with donate_argnames to avoid copies for z, v, phi_p, cos_phi_p
        k = jnp.arange(len(voxel_cylinder))
        z = gp.delta_voxel * (k - (num_slices - 1) / 2.0) + gp.recon_slice_offset  # recon_ijk_to_xyz
        v = pixel_mag * z  # geometry_xyz_to_uv_mag
        # Compute vertical cone angle of voxels
        phi_p = jnp.arctan2(v, gp.source_detector_dist)  # compute_vertical_data_single_pixel
        cos_phi_p = jnp.cos(phi_p)  # We assume the vertical angle |phi_p| < 45 degrees so cos_alpha_p_z = cos_phi_p
        scaled_voxel_values = voxel_cylinder / cos_phi_p
        # End TODO

        # Get the length of projection of detector on vertical voxel profile (in fraction of voxel size)
        # This is also the slope of the map from voxel index to detector index
        W_p_r = (pixel_mag * gp.delta_voxel) / gp.delta_det_row
        slope_k_to_m = W_p_r
        L_max = jnp.minimum(1, W_p_r)  # Maximum fraction of a detector that can be covered by one voxel.

        # Set up detector row indices array (0, 10, 20, ..., 10*num_slice_batches)
        det_rows_per_batch = gp.entries_per_cylinder_batch
        det_rows_per_batch = min(det_rows_per_batch, num_det_rows)
        num_det_row_batches = (num_det_rows + det_rows_per_batch - 1) // det_rows_per_batch
        det_row_indices = det_rows_per_batch * jnp.arange(num_det_row_batches)

        det_center_row = (num_det_rows - 1) / 2.0
        row_batch = jnp.arange(det_rows_per_batch)

        # Set up a function to map over subsets of detector rows
        def create_det_column_rows(start_index):
            # We need to match the back projector, so we have to determine the fraction of each voxel that projects
            # to each detector.
            # First project the detector centers to the voxel cylinder
            m_center = start_index + row_batch  # Center of detector elements
            v_m = (m_center - det_center_row) * gp.delta_det_row - gp.det_row_offset  # Detector center in ALUs
            z_m = v_m / pixel_mag  # z coordinate of the projection of the center of the first detector element in this batch
            # Convert to voxel fractional index and find the center of each voxel
            k_m = (z_m - gp.recon_slice_offset) / gp.delta_voxel + (num_slices - 1) / 2.0
            k_m_center = jnp.round(k_m).astype(int)  # Center of the voxel hit by the center of the detector
            # Then map the center of the voxels back to the detector.
            m_p = slope_k_to_m * (k_m_center - k_m[0]) + m_center[0]  # Projection to detector of voxel centers

            # Allocate space
            new_column_batch = jnp.zeros(det_rows_per_batch)
            # Do the vertical projection
            for k_offset in jnp.arange(start=-gp.bp_psf_radius, stop=gp.bp_psf_radius+1):
                k_ind = k_m_center + k_offset  # Indices of the current set of voxels touched by the detector elements
                # The projection of these centers is the projection of k_m_center (which is m_p) plus
                # the offset times the slope of the map from voxel index to detector index
                abs_delta_p_r_m = jnp.abs(m_p + slope_k_to_m * k_offset - m_center)  # Distance from projection of center of voxel to center of detector
                A_row_k = jnp.clip((W_p_r + 1) / 2 - abs_delta_p_r_m, 0, L_max)  # Fraction of the detector hit by this voxel
                A_row_k *= (k_ind >= 0) * (k_ind < num_slices)
                new_column_batch = jnp.add(new_column_batch, A_row_k * scaled_voxel_values[k_ind])

            return new_column_batch, None

        det_column, _ = jax.lax.map(create_det_column_rows, det_row_indices)
        det_column = det_column.flatten()
        det_column = jax.lax.slice_in_dim(det_column, 0, num_det_rows)
        return det_column

    @staticmethod
    @partial(jax.jit, static_argnames='projector_params')
    def back_project_one_view_to_pixel_batch(sinogram_view, pixel_indices, single_view_params, projector_params, coeff_power=1):
        """
        Use vmap to do a backprojection from one view to multiple pixels (voxel cylinders).

        Args:
            sinogram_view (2D jax array): one view of the sinogram to be back projected.
                2D jax array of shape (num_det_rows)x(num_det_channels)
            pixel_indices (1D jax array of int):  indices into flattened array of size num_rows x num_cols.
            single_view_params: These are the view dependent parameters for the view being back projected.
            projector_params (namedtuple): tuple of (sinogram_shape, recon_shape, get_geometry_params()).
            coeff_power (int): backproject using the coefficients of (A_ij ** coeff_power).
                Normally 1, but should be 2 when computing Hessian diagonal.

        Returns:
            The voxel values for all slices at the input index (i.e., a voxel cylinder) obtained by backprojecting
            the input sinogram view.
        """

        vertical_fan_projector = ConeBeamModel.back_vertical_fan_one_view_to_pixel_batch
        horizontal_fan_projector = ConeBeamModel.back_horizontal_fan_one_view_to_pixel_batch

        det_voxel_cylinder = horizontal_fan_projector(sinogram_view, pixel_indices, single_view_params,
                                                      projector_params, coeff_power=coeff_power)
        back_projection = vertical_fan_projector(det_voxel_cylinder, pixel_indices, single_view_params,
                                                 projector_params, coeff_power=coeff_power)

        return back_projection

    @staticmethod
    def back_horizontal_fan_one_view_to_pixel_batch(sinogram_view, pixel_indices, angle,
                                                    projector_params, coeff_power=1):
        """
        Apply the back projection of a horizontal fan beam transformation to a single sinogram view
        and return the resulting voxel cylinders.

        Args:
            sinogram_view (2D jax array): one view of the sinogram to be back projected.
                2D jax array of shape (num_det_rows)x(num_det_channels)
            pixel_indices (1D jax array of int):  indices into flattened array of size num_rows x num_cols.
            angle (float): The projection angle in radians for this view.
            projector_params (namedtuple): tuple of (sinogram_shape, recon_shape, get_geometry_params()).
            coeff_power (int): backproject using the coefficients of (A_ij ** coeff_power).
                Normally 1, but should be 2 when computing Hessian diagonal.
        Returns:
            jax array of shape (len(pixel_indices), num_det_rows)
        """
        # Get all the geometry parameters - we use gp since geometry parameters is a named tuple and we'll access
        # elements using, for example, gp.delta_det_channel, so a longer name would be clumsy.
        gp = projector_params.geometry_params

        num_views, num_det_rows, num_det_channels = projector_params.sinogram_shape
        num_pixels = pixel_indices.shape[0]

        # Get the data needed for horizontal projection
        n_p, n_p_center, W_p_c, cos_alpha_p_xy = ConeBeamModel.compute_horizontal_data(pixel_indices, angle, projector_params)
        L_max = jnp.minimum(1, W_p_c)

        # Allocate the voxel cylinder array
        det_voxel_cylinder = jnp.zeros((num_pixels, num_det_rows))

        # Do the horizontal projection
        for n_offset in jnp.arange(start=-gp.psf_radius, stop=gp.psf_radius + 1):
            n = n_p_center + n_offset
            abs_delta_p_c_n = jnp.abs(n_p - n)
            L_p_c_n = jnp.clip((W_p_c + 1) / 2 - abs_delta_p_c_n, 0, L_max)
            A_chan_n = gp.delta_voxel * L_p_c_n / cos_alpha_p_xy
            A_chan_n *= (n >= 0) * (n < num_det_channels)
            A_chan_n = A_chan_n ** coeff_power
            det_voxel_cylinder = jnp.add(det_voxel_cylinder, A_chan_n.reshape((-1, 1)) * sinogram_view[:, n].T)

        return det_voxel_cylinder

    @staticmethod
    def back_vertical_fan_one_view_to_pixel_batch(det_voxel_cylinder, pixel_indices, single_view_params,
                                                  projector_params, coeff_power=1):
        """
        Apply a fan beam backward projection in the vertical direction to the pixel determined by indices
        into the flattened array of size num_rows x num_cols.  This returns a vector obtained from the projection of
        the detector-based voxel cylinders onto voxel cylinders in recon space, so the output vector has length
        num_recon_slices.

        Args:
            det_voxel_cylinder (2D jax array): 2D array of shape (num_pixels, num_det_rows) of voxel values, where
                det_voxel_cylinder[i, j] is the value of the voxel in row j at the location determined by indices[i].
            pixel_indices (1D jax array of int):  indices into flattened array of size num_rows x num_cols.
            single_view_params: These are the view dependent parameters for the view being back projected.
            projector_params (namedtuple): tuple of (sinogram_shape, recon_shape, get_geometry_params()).
            coeff_power (int): backproject using the coefficients of (A_ij ** coeff_power).
                Normally 1, but should be 2 when computing Hessian diagonal.

        Returns:
            2D jax array of shape (num_pixels, num_recon_slices) of voxel values.
        """
        pixel_map = jax.vmap(ConeBeamModel.back_vertical_fan_one_view_to_one_pixel,
                             in_axes=(0, 0, None, None, None))
        new_pixels = pixel_map(det_voxel_cylinder, pixel_indices, single_view_params, projector_params, coeff_power)

        return new_pixels

    @staticmethod
    def back_vertical_fan_one_view_to_one_pixel(detector_column_values, pixel_index, angle, projector_params,
                                                coeff_power=1):
        """
        Apply the back projection of a vertical fan beam transformation to a single voxel cylinder and return the column
        vector of the resulting values.

        Args:
            detector_column_values (1D jax array): 1D array of shape (num_det_rows,) of voxel values, where
                detector_column_values[i, j] is the value of the voxel in row j at the location determined by indices[i].
            pixel_index (int):  Index into flattened array of size num_rows x num_cols.
            angle (float): The projection angle in radians for this view.
            projector_params (namedtuple): tuple of (sinogram_shape, recon_shape, get_geometry_params()).
            coeff_power (int): backproject using the coefficients of (A_ij ** coeff_power).
                Normally 1, but should be 2 when computing Hessian diagonal.

        Returns:
            1D jax array of shape (num_recon_slices,) of voxel values.
        """
        # Get all the geometry parameters - we use gp since geometry parameters is a named tuple and we'll access
        # elements using, for example, gp.delta_det_channel, so a longer name would be clumsy.
        gp = projector_params.geometry_params

        num_views, num_det_rows, num_det_channels = projector_params.sinogram_shape
        recon_shape = projector_params.recon_shape
        num_recon_rows, num_recon_cols, num_recon_slices = recon_shape

        # Set up slice indices array (0, slices_per_batch, 2*slices_per_batch, ..., num_slice_batches*slices_per_batch)
        slices_per_batch = gp.entries_per_cylinder_batch
        slices_per_batch = min(slices_per_batch, num_recon_slices)
        num_slice_batches = (num_recon_slices + slices_per_batch - 1) // slices_per_batch
        slice_indices = slices_per_batch * jnp.arange(num_slice_batches)

        # Set up a function to map over the slices of the cylinder
        # Here we can use a map over subsections of the voxel cylinder because we are indexing by slice,
        # so there is no overlap from one section to the next.
        def create_voxel_cylinder_slices(start_index):
            # Allocate space
            new_cylinder = jnp.zeros(slices_per_batch)
            # Get the data needed for vertical projection
            cur_slice_indices = start_index + jnp.arange(slices_per_batch)
            m_p, m_p_center, W_p_r, cos_alpha_p_z = ConeBeamModel.compute_vertical_data_single_pixel(pixel_index, cur_slice_indices, angle,
                                                                                                     projector_params)
            L_max = jnp.minimum(1, W_p_r)  # Maximum fraction of a detector that can be covered by one voxel.

            # Do the vertical projection
            for m_offset in jnp.arange(start=-gp.psf_radius, stop=gp.psf_radius+1):
                m = m_p_center + m_offset
                abs_delta_p_r_m = jnp.abs(m_p - m)  # Distance from projection of center of voxel to center of detector
                L_p_r_m = jnp.clip((W_p_r + 1) / 2 - abs_delta_p_r_m, 0, L_max)
                A_row_m = L_p_r_m / cos_alpha_p_z
                A_row_m *= (m >= 0) * (m < num_det_rows)
                A_row_m = A_row_m ** coeff_power
                new_cylinder = jnp.add(new_cylinder, A_row_m * detector_column_values[m])

            return new_cylinder, None

        recon_voxel_cylinder, _ = jax.lax.map(create_voxel_cylinder_slices, slice_indices)
        recon_voxel_cylinder = recon_voxel_cylinder.flatten()
        recon_voxel_cylinder = jax.lax.slice_in_dim(recon_voxel_cylinder, 0, num_recon_slices)
        return recon_voxel_cylinder

    @staticmethod
    def compute_vertical_data_single_pixel(pixel_index, slice_indices, angle, projector_params):
        """
        Compute the quantities m_p, m_p_center, W_p_r, cos_alpha_p_z needed for vertical projection.

        Args:
            pixel_index (int):  Index into flattened array of size num_rows x num_cols.
            slice_indices (array of int): Indices into the recon slices.
            angle (float): The projection angle in radians for this view.
            projector_params (namedtuple): tuple of (sinogram_shape, recon_shape, get_geometry_params()).

        Returns:
            m_p, m_p_center, W_p_r, cos_alpha_p_z
        """
        # Get all the geometry parameters - we use gp since geometry parameters is a named tuple and we'll access
        # elements using, for example, gp.delta_det_channel, so a longer name would be clumsy.
        gp = projector_params.geometry_params

        num_views, num_det_rows, num_det_channels = projector_params.sinogram_shape
        recon_shape = projector_params.recon_shape
        num_recon_rows, num_recon_cols, num_recon_slices = recon_shape

        # Convert the index into (i,j,k) coordinates corresponding to the indices into the 3D voxel array
        row_index, col_index = jnp.unravel_index(pixel_index, recon_shape[:2])
        # slice_indices = jnp.arange(num_recon_slices)

        x_p, y_p, z_p = ConeBeamModel.recon_ijk_to_xyz(row_index, col_index, slice_indices, gp.delta_voxel,
                                                       recon_shape, gp.recon_slice_offset, angle)

        # Convert from xyz to coordinates on detector
        u_p, v_p, pixel_mag = ConeBeamModel.geometry_xyz_to_uv_mag(x_p, y_p, z_p, gp.source_detector_dist, gp.magnification)
        # Convert from uv to index coordinates in detector and get the vector of center detector rows for this cylinder
        m_p, _ = ConeBeamModel.detector_uv_to_mn(u_p, v_p, gp.delta_det_channel, gp.delta_det_row, gp.det_channel_offset,
                                                 gp.det_row_offset, num_det_rows, num_det_channels)
        m_p_center = jnp.round(m_p).astype(int)

        # Compute vertical cone angle of pixel
        phi_p = jnp.arctan2(v_p, gp.source_detector_dist)

        # Compute cos alpha for row and columns
        cos_phi_p = jnp.cos(phi_p)  # We assume the vertical angle |phi_p| < 45 degrees so cos_alpha_p_z = cos_phi_p
        # cos_alpha_p_z = jnp.maximum(jnp.abs(cos_phi_p), jnp.abs(jnp.sin(phi_p)))

        # Get the length of projection of flattened voxel on detector (in fraction of detector size)
        W_p_r = pixel_mag * (gp.delta_voxel / gp.delta_det_row)   # * cos_alpha_p_z / cos_phi_p

        vertical_data = (m_p, m_p_center, W_p_r, cos_phi_p)  # cos_alpha_p_z)

        return vertical_data

    @staticmethod
    def compute_horizontal_data(pixel_indices, angle, projector_params):
        """
        Compute the quantities n_p, n_p_center, W_p_c, cos_alpha_p_xy needed for vertical projection.

        Args:
            pixel_indices (1D jax array of int):  indices into flattened array of size num_rows x num_cols.
            angle (float): The projection angle in radians for this view.
            projector_params (namedtuple): tuple of (sinogram_shape, recon_shape, get_geometry_params()).

        Returns:
            n_p, n_p_center, W_p_c, cos_alpha_p_xy
        """
        # Get all the geometry parameters - we use gp since geometry parameters is a named tuple and we'll access
        # elements using, for example, gp.delta_det_channel, so a longer name would be clumsy.
        gp = projector_params.geometry_params

        num_views, num_det_rows, num_det_channels = projector_params.sinogram_shape
        recon_shape = projector_params.recon_shape
        num_recon_rows, num_recon_cols, num_recon_slices = recon_shape

        # Convert the index into (i,j,k) coordinates corresponding to the indices into the 3D voxel array
        row_index, col_index = jnp.unravel_index(pixel_indices, recon_shape[:2])
        slice_index = jnp.arange(1)

        x_p, y_p, _ = ConeBeamModel.recon_ijk_to_xyz(row_index, col_index, slice_index, gp.delta_voxel,
                                                     recon_shape, gp.recon_slice_offset, angle)

        # Convert from xyz to coordinates on detector
        # pixel_mag should be kept in terms of magnification to allow for source_detector_dist = jnp.Inf
        pixel_mag = 1 / (1 / gp.magnification - y_p / gp.source_detector_dist)
        # Compute the physical position that this voxel projects onto the detector
        u_p = pixel_mag * x_p
        det_center_channel = (num_det_channels - 1) / 2.0  # num_of_cols

        # Calculate indices on the detector grid
        n_p = (u_p + gp.det_channel_offset) / gp.delta_det_channel + det_center_channel  # Sync with detector_uv_to_mn
        n_p_center = jnp.round(n_p).astype(int)

        # Compute horizontal and vertical cone angle of pixel
        theta_p = jnp.arctan2(u_p, gp.source_detector_dist)

        # Compute cos alpha for row and columns
        cos_alpha_p_xy = jnp.maximum(jnp.abs(jnp.cos(angle - theta_p)),
                                    jnp.abs(jnp.sin(angle - theta_p)))

        # Compute projected voxel width along columns and rows (in fraction of detector size)
        W_p_c = pixel_mag * (gp.delta_voxel / gp.delta_det_channel) * (cos_alpha_p_xy / jnp.cos(theta_p))

        horizontal_data = (n_p, n_p_center, W_p_c, cos_alpha_p_xy)

        return horizontal_data

    @staticmethod
    def recon_ijk_to_xyz(i, j, k, delta_voxel, recon_shape, recon_slice_offset, angle):
        """
        Convert (i, j, k) indices into the recon volume to corresponding (x, y, z) coordinates.
        """
        num_recon_rows, num_recon_cols, num_recon_slices = recon_shape

        # Compute the un-rotated coordinates relative to iso
        # Note the change in order from (i, j) to (y, x)!!
        y_tilde = delta_voxel * (i - (num_recon_rows - 1) / 2.0)
        x_tilde = delta_voxel * (j - (num_recon_cols - 1) / 2.0)

        # Precompute cosine and sine of view angle, then do the rotation
        cosine = jnp.cos(angle)  # length = num_views
        sine = jnp.sin(angle)  # length = num_views

        x = cosine * x_tilde - sine * y_tilde
        y = sine * x_tilde + cosine * y_tilde

        z = delta_voxel * (k - (num_recon_slices - 1) / 2.0) + recon_slice_offset
        return x, y, z

    @staticmethod
    def geometry_xyz_to_uv_mag(x, y, z, source_detector_dist, magnification):
        """
        Convert (x, y, z) coordinates to to (u, v) detector coordinates plus the pixel-dependent magnification.
        """
        # Compute the magnification at this specific voxel
        # The following expression is valid even when source_detector_dist = jnp.Inf
        pixel_mag = 1 / (1 / magnification - y / source_detector_dist)

        # Compute the physical position that this voxel projects onto the detector
        u = pixel_mag * x
        v = pixel_mag * z

        return u, v, pixel_mag

    @staticmethod
    @jax.jit
    def detector_uv_to_mn(u, v, delta_det_channel, delta_det_row, det_channel_offset, det_row_offset, num_det_rows,
                          num_det_channels):
        """
        Convert (u, v) detector coordinates to fractional indices (m, n) into the detector.

        Note:
            This version does not account for nonzero detector rotation.
        """
        # Account for small rotation of the detector
        # TODO:  In addition to including the rotation, we'd need to adjust the calculation of the channel as a
        #  function of slice.
        u_tilde = u  # jnp.cos(det_rotation) * u + jnp.sin(det_rotation) * v
        v_tilde = v  # -jnp.sin(det_rotation) * u + jnp.cos(det_rotation) * v

        # Get the center of the detector grid for columns and rows
        det_center_row = (num_det_rows - 1) / 2.0  # num_of_rows
        det_center_channel = (num_det_channels - 1) / 2.0  # num_of_cols

        # Calculate indices on the detector grid
        m = (v_tilde + det_row_offset) / delta_det_row + det_center_row
        n = (u_tilde + det_channel_offset) / delta_det_channel + det_center_channel  # Sync with compute_horizontal_data

        return m, n

    @staticmethod
    @jax.jit
    def detector_mn_to_uv(m, n, delta_det_channel, delta_det_row, det_channel_offset, det_row_offset, num_det_rows,
                      num_det_channels):
        """
        Convert fractional detector grid indices (m, n) into detector coordinates (u, v).

        Parameters:
            m: Fractional row index on the detector grid (vertical direction).
            n: Fractional channel index on the detector grid (horizontal direction).
            delta_det_channel: Spacing (pitch) of the detector channels (horizontal direction).
            delta_det_row: Spacing (pitch) of the detector rows (vertical direction).
            det_channel_offset: Offset in the detector channel (horizontal) direction.
            det_row_offset: Offset in the detector row (vertical) direction.
            num_det_rows: Total number of rows in the detector.
            num_det_channels: Total number of channels in the detector.

        Returns:
            u: Physical detector coordinate in the channel direction.
            v: Physical detector coordinate in the row direction.
        """
        # Calculate the center of the detector grid
        det_center_row = (num_det_rows - 1) / 2.0
        det_center_channel = (num_det_channels - 1) / 2.0

        # Compute detector coordinates (u, v)
        v = (m - det_center_row) * delta_det_row - det_row_offset
        u = (n - det_center_channel) * delta_det_channel - det_channel_offset

        return u, v

    @staticmethod
    @jax.jit
    def compute_y_mag_for_pixel(pixel_index, angle, recon_shape, projector_params):

        gp = projector_params.geometry_params
        row_index, col_index = jnp.unravel_index(pixel_index, recon_shape[:2])

        # Compute the un-rotated coordinates relative to iso
        # Note the change in order from (i, j) to (y, x)!!
        y_tilde = gp.delta_voxel * (row_index - (recon_shape[0] - 1) / 2.0)
        x_tilde = gp.delta_voxel * (col_index - (recon_shape[1] - 1) / 2.0)

        # Precompute cosine and sine of view angle, then do the rotation
        cosine = jnp.cos(angle)  # length = num_views
        sine = jnp.sin(angle)  # length = num_views

        y = sine * x_tilde + cosine * y_tilde

        # Convert from xyz to coordinates on detector
        pixel_mag = 1 / (1 / gp.magnification - y / gp.source_detector_dist)
        return y, pixel_mag

    def direct_recon(self, sinogram, filter_name="ramp", view_batch_size=DIRECT_RECON_VIEW_BATCH_SIZE):
        return self.fdk_recon(sinogram, filter_name=filter_name, view_batch_size=view_batch_size)

    def direct_filter(self, sinogram, filter_name="ramp", view_batch_size=DIRECT_RECON_VIEW_BATCH_SIZE):
        """
        Perform filtering on the given sinogram as needed for an FBP/FDK or other direct recon.

        Args:
            sinogram (jax array): The input sinogram with shape (num_views, num_rows, num_channels).
            filter_name (string, optional): Name of the filter to be used. Defaults to "ramp"
            view_batch_size (int, optional):  Size of view batches (used to limit memory use)

        Returns:
            filtered_sinogram (jax array): The sinogram after FBP filtering.
        """
        return self.fdk_filter(sinogram, filter_name=filter_name, view_batch_size=view_batch_size)

    def fdk_filter(self, sinogram, filter_name="ramp", view_batch_size=DIRECT_RECON_VIEW_BATCH_SIZE):
        """
        Perform FDK filtering on the given sinogram.

        Args:
            sinogram (jax array): The input sinogram with shape (num_views, num_rows, num_channels).
            filter_name (string, optional): Name of the filter to be used. Defaults to "ramp"
            view_batch_size (int, optional):  Size of view batches (used to limit memory use)

        Returns:
            filtered_sinogram (jax array): The sinogram after FDK filtering.
        """
        # Get parameters
        num_views, num_rows, num_channels = sinogram.shape
        source_detector_dist, source_iso_dist = self.get_params(['source_detector_dist', 'source_iso_dist'])
        delta_voxel, delta_det_row, delta_det_channel = self.get_params(['delta_voxel', 'delta_det_row', 'delta_det_channel'])
        det_row_offset, det_channel_offset = self.get_params(['det_row_offset', 'det_channel_offset'])

        if view_batch_size is None:
            view_batch_size = self.view_batch_size_for_vmap
            max_view_batch_size = 128  # Limit the view batch size here and ParallelBeam due to https://github.com/jax-ml/jax/issues/27591
            view_batch_size = min(view_batch_size, max_view_batch_size)

        # Magnification factor M_0 = Source-Detector Distance / Source-Isocenter Distance
        M_0 = self.get_magnification()

        # Define the index arrays for channels and rows
        m = jnp.arange(num_rows)  # Column vector for rows
        n = jnp.arange(num_channels)  # Row vector for channels
        m_grid, n_grid = jnp.meshgrid(m, n, indexing='ij')

        # Coordinate transformation to physical distances:
        u_grid, v_grid = self.detector_mn_to_uv(m_grid, n_grid, delta_det_channel, delta_det_row,
                                                det_channel_offset, det_row_offset, num_rows, num_channels)

        # Compute the weight
        weight_map = source_detector_dist / jnp.sqrt(source_detector_dist ** 2 + u_grid**2 + v_grid**2)

        # Apply the pre-weighting factor to the sinogram
        weighted_sinogram = jax.device_put(sinogram * weight_map[None, :, :], self.sinogram_device)

        # Compute the scaled filter
        # Scaling factor alpha adjusts the filter to account for voxel size, ensuring consistent reconstruction.
        # For a detailed theoretical derivation of this scaling factor, please refer to the zip file linked at
        # https://mbirjax.readthedocs.io/en/latest/theory.html
        recon_filter = tomography_utils.generate_direct_recon_filter(num_channels, filter_name=filter_name)
        alpha = delta_det_row / (delta_voxel**3 * M_0)
        recon_filter = alpha * recon_filter

        # Define convolution for a single row (across its channels)
        def convolve_row(row):
            return jax.scipy.signal.fftconvolve(row, recon_filter, mode="valid")

        # Apply above convolve func across each row of a view
        def apply_convolution_to_view(view):
            return jax.vmap(convolve_row)(view)

        # Apply convolution across the channels of the weighted sinogram per each fixed view & row
        num_views = sinogram.shape[0]
        filtered_sino_list = []
        for i in range(0, num_views, view_batch_size):
            sino_batch = jax.device_put(weighted_sinogram[i:min(i + view_batch_size, num_views)], self.worker)
            filtered_sinogram_batch = jax.lax.map(apply_convolution_to_view, sino_batch, batch_size=view_batch_size)
            filtered_sinogram_batch.block_until_ready()
            filtered_sino_list.append(jax.device_put(filtered_sinogram_batch, self.sinogram_device))
        filtered_sinogram = jnp.concatenate(filtered_sino_list, axis=0)
        filtered_sinogram *= jnp.pi / num_views
        return filtered_sinogram

    def fdk_recon(self, sinogram, filter_name="ramp", view_batch_size=DIRECT_RECON_VIEW_BATCH_SIZE):
        """
        Perform FDK reconstruction on the given sinogram.

        Our implementation uses standard filtering of the sinogram, then uses the adjoint of the forward projector to
        perform the backprojection.  This is different from many implementations, in which the backprojection is not
        exactly the adjoint of the forward projection.  For a detailed theoretical derivation of this implementation,
        see the zip file linked at this page: https://mbirjax.readthedocs.io/en/latest/theory.html

        Args:
            sinogram (jax array): The input sinogram with shape (num_views, num_rows, num_channels).
            filter_name (string, optional): Name of the filter to be used. Defaults to "ramp"
            view_batch_size (int, optional):  Size of view batches (used to limit memory use)

        Returns:
            recon (jax array): The reconstructed volume after FDK reconstruction.
        """

        filtered_sinogram = self.fdk_filter(sinogram, filter_name=filter_name, view_batch_size=view_batch_size)

        # Apply backprojection
        recon = self.back_project(filtered_sinogram)

        return recon

    def split_sino_recon(self, sino, weights=None, half_overlap=5, init_recon=None, max_iterations=15, stop_threshold_change_pct=0.2,
                         first_iteration=0, compute_prior_loss=False, logfile_path='./logs/recon.log', print_logs=True):
        """
        This function reduces memory usage for cone beam MBIR reconstruction by approximately a factor of 2
        by splitting the detector rows into two overlapping halves, reconstructing each half separately,
        and stitching the reconstructions together.

        The function can be called with the same arguments as TomographyModel.recon(), and it should return a
        reconstruction which is approximately equal to the reconstruction returned by TomographyModel.recon().

        Args:
            sino (jnp.ndarray | np.ndarray): Full sinogram of shape (num_views, num_rows, num_cols).
            weights (jnp.ndarray | np.ndarray, optional): Optional sinogram weights with the same shape as `sino`.
            half_overlap (int): Number of overlapping detector rows and recon slices per half. (total overlap = 2 * half_overlap)
            init_recon (optional): Same as in the recon method.
            max_iterations (int, optional): Same as in the recon method.
            stop_threshold_change_pct (float, optional): Same as in the recon method.
            first_iteration (int, optional): Same as in the TomographyModel.recon() method.
            compute_prior_loss (bool, optional): Same as in the TomographyModel.recon() method.
            logfile_path (str, optional): Same as in the TomographyModel.recon() method.
            print_logs (bool, optional): Same as in the TomographyModel.recon() method.

        Returns:
            Tuple[jnp.ndarray, dict]:
                - Reconstructed volume (jax array).
                - Dictionary of metadata containing recon and model parameters for each half.

        Raises:
            ValueError: If inputs are missing or shapes are inconsistent.
            AssertionError: If array dimensions are invalid.
            TypeError: If half_overlap is not an integer.

        Example:
            >>> import jax.numpy as jnp
            >>> import mbirjax as mj
            >>> sino = jnp.ones((180, 64, 64))  # (views, rows, cols)
            >>> model = mj.ConeBeamModel(sinogram_shape=sino.shape,
            ...                          angles=jnp.linspace(0, jnp.pi, 180),
            ...                          source_detector_dist=1000.0,
            ...                          source_iso_dist=500.0)
            >>> recon, recon_info = model.split_sino_recon(sino, half_overlap=4)
            >>> recon.shape  # Quilted reconstruction volume
            (64, 64, 64)
        """
        # -------- Basic validation --------
        if sino is None:
            raise ValueError("sino must be provided.")
        if not (hasattr(sino, "ndim") and sino.ndim == 3):
            raise AssertionError("sino must be a 3D array shaped (num_views, num_rows, num_cols).")
        if weights is not None and getattr(weights, "shape", None) != sino.shape:
            raise AssertionError("weights, if provided, must have the same shape as sino.")

        # Get parameters for later use
        num_views, full_num_rows, num_cols = sino.shape

        # -------- parameters needed to create top and bottom models --------
        delta_det_row = self.get_params('delta_det_row')
        full_det_row_offset = self.get_params('det_row_offset')
        delta_voxel = self.get_params('delta_voxel')
        full_recon_shape = self.get_params('recon_shape')
        full_recon_slice_offset = self.get_params('recon_slice_offset')
        magnification = self.get_magnification()

        # Compute overlaps for sinogram and recon
        delta_detector_row_at_iso = max(delta_det_row / magnification, 1e-12)
        ratio_pixel_to_sino_pitch = delta_voxel / delta_detector_row_at_iso
        if ratio_pixel_to_sino_pitch > 1:
            half_overlap_sino = int(jnp.round(half_overlap * ratio_pixel_to_sino_pitch))
            half_overlap_recon = half_overlap
        else:
            half_overlap_sino = half_overlap
            half_overlap_recon = int(jnp.round(half_overlap * 1/ratio_pixel_to_sino_pitch))

        """
        Compute detector shape parameters for top and bottom sinograms
        """
        # -------- Choose the detector row nearest to iso --------
        det_iso_row_float = ((full_num_rows - 1) / 2.0) + (full_det_row_offset / delta_det_row)
        det_iso_row_index = int(jnp.round(det_iso_row_float))

        # Validate iso-row index is inside (0, num_rows)
        if not (0 < det_iso_row_index < full_num_rows):
            raise ValueError(
                f"Computed det_iso_row_index={det_iso_row_index} is out of valid range (0, {full_num_rows-1}). "
            )

        # -------- Detector row ranges for top and bottom sinogram halves --------
        top_lo = 0
        top_hi = min(det_iso_row_index + half_overlap_sino, full_num_rows)
        bot_lo = max(det_iso_row_index - half_overlap_sino, 0)
        bot_hi = full_num_rows

        # -------- Split sinogram (and weights) into top and bottom halves --------
        sino_top_half = sino[:, top_lo:top_hi, :]
        sino_bot_half = sino[:, bot_lo:bot_hi, :]

        weights_top_half = None
        weights_bot_half = None
        if weights is not None:
            weights_top_half = weights[:, top_lo:top_hi, :]
            weights_bot_half = weights[:, bot_lo:bot_hi, :]

        # -------- Calculate number of rows and center location for top and bottom sinograms --------
        top_num_rows = top_hi - top_lo
        bot_num_rows = bot_hi - bot_lo

        full_det_center = (full_num_rows - 1) / 2.0
        top_det_center = (top_num_rows - 1) / 2.0
        bot_det_center = (bot_num_rows - 1) / 2.0

        # -------- Calculate detector row offsets required for top and bottom models --------
        top_det_row_offset = full_det_row_offset + (full_det_center - (top_det_center + top_lo)) * delta_det_row
        bot_det_row_offset = full_det_row_offset + (full_det_center - (bot_det_center + bot_lo)) * delta_det_row

        # Set the regularization parameters from the full sinogram
        self.auto_set_regularization_params(sino)

        # -------- Build top-half model --------
        ct_model_top_half = mj.copy_ct_model(self, new_num_det_rows=top_num_rows)
        ct_model_top_half.set_params(det_row_offset=top_det_row_offset)
        ct_model_top_half.set_params(auto_regularize_flag=False)

        # -------- Build bottom-half model --------
        ct_model_bot_half = mj.copy_ct_model(self, new_num_det_rows=bot_num_rows)
        ct_model_bot_half.set_params(det_row_offset=bot_det_row_offset)
        ct_model_bot_half.set_params(auto_regularize_flag=False)

        """
        Compute recon shape parameters for top and bottom reconstructions
        """
        # Get recon shape parameters for later use
        full_recon_rows, full_recon_cols, full_recon_slices = full_recon_shape

        # -------- Compute the recon slice nearest to iso --------
        full_recon_iso_slice_index_float = (full_recon_slices - 1) / 2.0 - full_recon_slice_offset/delta_voxel
        split_index = int(jnp.round(full_recon_iso_slice_index_float))
        top_num_slices = split_index + 1

        # Compute the offset of the split from iso.
        # This will be used to slightly shift the slices so that they align with a standard reconstruction.
        split_offset = split_index - full_recon_iso_slice_index_float

        # Fallback: If split index creates an empty top or bottom half sinogram, then warn and do a normal MBIR recon.
        if (split_index < 1) or (split_index > full_recon_slices - 2):
            warnings.warn(
                "split_index is too close to the volume boundary; falling back to standard MBIR reconstruction.",
                UserWarning,
            )
            return self.recon(
                sino,
                weights=weights,
                init_recon=init_recon,
                max_iterations=max_iterations,
                stop_threshold_change_pct=stop_threshold_change_pct,
                first_iteration=first_iteration,
                compute_prior_loss=compute_prior_loss,
                logfile_path=logfile_path,
                print_logs=print_logs,
            )

        # -------- Compute and set the shapes of top and bottom recons --------
        top_recon_shape = (full_recon_shape[0], full_recon_shape[1], top_num_slices + half_overlap_recon)
        bot_recon_shape = (full_recon_shape[0], full_recon_shape[1], (full_recon_shape[2] - top_num_slices) + half_overlap_recon)

        ct_model_top_half.set_params(recon_shape=top_recon_shape)
        ct_model_bot_half.set_params(recon_shape=bot_recon_shape)

        # -------- Compute and set the offsets of top and bottom recons --------
        top_recon_slice_offset = (+half_overlap_recon - (top_recon_shape[2]-1)/2 + 0 + split_offset) * delta_voxel
        bot_recon_slice_offset = (-half_overlap_recon + (bot_recon_shape[2]-1)/2 + 1 + split_offset) * delta_voxel

        ct_model_top_half.set_params(recon_slice_offset=top_recon_slice_offset)
        ct_model_bot_half.set_params(recon_slice_offset=bot_recon_slice_offset)

        # -------- Reconstruct halves (pass weights if provided) --------
        recon_top_half, recon_top_dict = ct_model_top_half.recon(sino_top_half, weights=weights_top_half,
                                                                 init_recon=init_recon, max_iterations=max_iterations,
                                                                 stop_threshold_change_pct=stop_threshold_change_pct,
                                                                 first_iteration=first_iteration,
                                                                 compute_prior_loss=compute_prior_loss,
                                                                 logfile_path=logfile_path, print_logs=print_logs)
        recon_bot_half, recon_bot_dict = ct_model_bot_half.recon(sino_bot_half, weights=weights_bot_half,
                                                                 init_recon=init_recon, max_iterations=max_iterations,
                                                                 stop_threshold_change_pct=stop_threshold_change_pct,
                                                                 first_iteration=first_iteration,
                                                                 compute_prior_loss=compute_prior_loss,
                                                                 logfile_path=logfile_path, print_logs=print_logs)

        # -------- Stitch together top and bottom reconstructions --------
        recon_full = mj.stitch_arrays([recon_top_half, recon_bot_half], overlap=2 * half_overlap_recon, axis=2)

        # -------- Construct full reconstruction dictionary --------
        recon_full_dict = {'recon_params_top': recon_top_dict['recon_params'],
                           'recon_params_bottom': recon_bot_dict['recon_params'],
                           'recon_log_top': recon_top_dict['recon_log'],
                           'recon_log_bottom': recon_bot_dict['recon_log'],
                           'notes_top': recon_top_dict['notes'],
                           'notes_bottom': recon_bot_dict['notes'],
                           'model_params_top': recon_top_dict['model_params'],
                           'model_params_bottom': recon_bot_dict['model_params'], }

        return recon_full, recon_full_dict




import sys
import mbirjax as mj
import mbirjax.preprocess as mjp
import numpy as np

def get_experiment_params(experiment_name):
    """Returns experiment-specific parameters."""
    base_config = {
        "source_det_dist_mm": 190,
        "source_iso_dist_mm": 70,
        "det_pixel_pitch_mm": 75 / 1000,
        "x_view_space_mm": 11.4,
        "z_view_space_mm": 14,
        "num_x_translations": 5,
        "num_z_translations": 3,
        "num_det_rows": 1944,
        "num_det_channels": 3072,
        "num_pixels_in_object": 3,
        "phantom_type": "text",
        "text": ['P', 'P', 'P'],
        "object_width_mm": 22,
        "object_thickness_mm": 2.15,
        "sharpness": 1.0,
        "max_iterations": 15,
    }

    if experiment_name == "experiment1":
        return base_config

    elif experiment_name == "experiment2":
        override_config = {
            # Modify only what's different for experiment2
            # Example: "phantom_type": "dots",
        }
        return {**base_config, **override_config}

    else:
        raise ValueError(f"Unknown experiment: {experiment_name}")

def main():
    experiment = sys.argv[1] if len(sys.argv) > 1 else "experiment1"
    params = get_experiment_params(experiment)

    # Set parameters for experiment
    source_det_dist_mm = params["source_det_dist_mm"]
    source_iso_dist_mm = params["source_iso_dist_mm"]
    det_pixel_pitch_mm = params["det_pixel_pitch_mm"]
    x_view_space_mm = params["x_view_space_mm"]
    z_view_space_mm = params["z_view_space_mm"]
    num_x_translations = params["num_x_translations"]
    num_z_translations = params["num_z_translations"]
    num_det_rows = params["num_det_rows"]
    num_det_channels = params["num_det_channels"]
    num_pixels_in_object = params["num_pixels_in_object"]
    phantom_type = params["phantom_type"]
    text = params["text"]
    object_width_mm = params["object_width_mm"]
    object_thickness_mm = params["object_thickness_mm"]
    sharpness = params["sharpness"]
    max_iterations = params["max_iterations"]


    # Calculate physical parameters in ALU
    # Note: 1 ALU = detector pixel pitch at iso
    ALU_per_mm = source_det_dist_mm / (source_iso_dist_mm * det_pixel_pitch_mm)
    source_iso_dist_ALU = source_iso_dist_mm * ALU_per_mm
    source_det_dist_ALU = source_iso_dist_ALU
    x_view_spacing_ALU = x_view_space_mm * ALU_per_mm
    z_view_spacing_ALU = z_view_space_mm * ALU_per_mm
    half_angle_rad = np.arctan2(max(num_det_rows, num_det_channels) / 2.0, source_iso_dist_ALU)
    object_width_ALU = object_width_mm * ALU_per_mm
    object_thickness_ALU = object_thickness_mm * ALU_per_mm
    delta_recon_row = object_thickness_ALU / num_pixels_in_object

    # Generate translation vectors
    translation_vectors = mj.gen_translation_vectors(num_x_translations, num_z_translations, x_view_spacing_ALU, z_view_spacing_ALU)

    # Compute sinogram shape
    sino_shape = (translation_vectors.shape[0], num_det_rows, num_det_channels)

    # Initialize model for reconstruction.
    tct_model = mj.TranslationModel(sino_shape, translation_vectors, source_detector_dist=source_det_dist_ALU, source_iso_dist=source_iso_dist_ALU)
    tct_model.set_params(sharpness=sharpness)
    recon_shape = tct_model.get_params('recon_shape')

    # Set number of rows in recon to match desired object thickness
    recon_shape = (3*int(object_thickness_ALU / delta_recon_row), recon_shape[1], recon_shape[2])
    tct_model.set_params(recon_shape=recon_shape)
    tct_model.set_params(delta_recon_row=delta_recon_row)
    tct_model.set_params(qggmrf_nbr_wts=[1.0, 1.0, 0.1])


    print("\n*************** Experimental Parameters ***************")
    print("source_detector_dist (ALU):", tct_model.get_params('source_detector_dist'))
    print("source_iso_dist (ALU):", tct_model.get_params('source_iso_dist'))
    print("delta_det_channel (ALU):", tct_model.get_params('delta_det_channel'))
    print("delta_det_row (ALU):", tct_model.get_params('delta_det_row'))
    print("Magnification:", tct_model.get_magnification())
    print("Half cone angle (deg):", np.rad2deg(half_angle_rad))

    print("x,z view spacing (ALU):", x_view_spacing_ALU, z_view_spacing_ALU)
    print("Object width (ALU):", object_width_ALU)
    print("Object thickness (ALU):", object_thickness_ALU)

    print("Recon shape:", recon_shape)
    print("delta_voxel (ALU):", tct_model.get_params('delta_voxel'))
    print("delta_recon_row (ALU):", tct_model.get_params('delta_recon_row'))
    print("************************************************")


    # Print model parameters and display translation array
    #tct_model.print_params()
    mj.display_translation_vectors(translation_vectors, recon_shape)

    # Generate ground truth phantom
    text_start_index = (recon_shape[0] - num_pixels_in_object) // 2
    row_indices = list(range(text_start_index, text_start_index + num_pixels_in_object))
    gt_recon = mj.gen_translation_phantom(recon_shape=recon_shape, option=phantom_type, text=text,
                                          font_size=object_width_ALU,
                                          text_row_indices=row_indices)

    # View test sample
    mj.slice_viewer(gt_recon.transpose(0, 2, 1), title='Ground Truth Recon', slice_label='View', slice_axis=0)

    # Generate synthetic sonogram data
    sino = tct_model.forward_project(gt_recon)

    # View sinogram
    mj.slice_viewer(sino, slice_axis=0, vmin=0, vmax=1, title='Original sinogram', slice_label='View')

    # Perform MBIR reconstruction
    recon, recon_params = tct_model.recon(sino, max_iterations=max_iterations)

    # Display Results
    mj.slice_viewer(gt_recon.transpose(0, 2, 1), recon.transpose(0, 2, 1), vmin=0, vmax=1,
                    title='Object (left) vs. MBIR reconstruction (right)', slice_axis=0)

    # Save as animated gifs
    mj.save_volume_as_gif(gt_recon, "gt_recon.gif", vmin=0, vmax=1)
    mj.save_volume_as_gif(recon, "mbir_recon.gif", vmin=0, vmax=1)


if __name__ == '__main__':
    main()



import numpy as np
import os
import matplotlib.pyplot as plt
import mbirjax

param_dict = {    # Set parameters
    'num_views' : 128,
    'num_det_rows' : 40,
    'num_det_channels' : 256,
    'start_angle' : 0,
    'end_angle' : np.pi,
    'granularity': [1, 8, 48, 128, 256, 256, 256, 256, 256, ],
    'partition_sequence' : [0, 1, 2, 3, 4, 4, 4, 4, 4, 4, 4, 4,],
    'max_iterations' : 10
}


def display_images_for_abstract(image1, image2, image3, labels, fig_title=None, vmin=None, vmax=None,
                                cmap='gray', show_colorbar=False):
    # Set global font size
    plt.rcParams.update({'font.size': 15})  # Adjust font size here

    if vmin is None:
        vmin = 0.0
    if vmax is None:
        vmax = 1.0

    fig, ax = plt.subplots(nrows=1, ncols=3, figsize=(20, 6))
    fig.suptitle(fig_title)

    a0 = ax[0].imshow(image1, vmin=vmin, vmax=vmax, cmap=cmap)
    #plt.colorbar(a0, ax=ax[0])
    ax[0].set_title(labels[0])

    a0 = ax[1].imshow(image2, vmin=vmin, vmax=vmax, cmap=cmap)
    #plt.colorbar(a0, ax=ax[1])
    ax[1].set_title(labels[1])

    a2 = ax[2].imshow(image3, vmin=vmin, vmax=vmax, cmap=cmap)
    #plt.colorbar(a2, ax=ax[2])
    ax[2].set_title(labels[2])

    if show_colorbar:
        fig.subplots_adjust(right=0.8)
        cbar_ax = fig.add_axes((0.81, 0.16, 0.02, 0.68))
        fig.colorbar(a2, cax=cbar_ax)

    plt.show()
    figure_folder_name = mj.make_figure_folder()
    os.makedirs(figure_folder_name, exist_ok=True)
    fig.savefig(os.path.join(figure_folder_name, fig_title + '.png'), bbox_inches='tight')




import mbirjax as mj
import numpy as np
import jax.numpy as jnp
import time
import sys
import jax


def evaluate_over_indices(filename, nv, nc, nr):

    if mj.get_memory_stats(print_results=False) is None:
        raise EnvironmentError('This script is for gpu only.')

    print('Loading {}'.format(filename))
    # Load the existing data
    data = np.load(filename, allow_pickle=True)
    mem_values = data['mem_values']
    time_values = data['time_values']
    eval_type_index = data['eval_type_index']
    max_percent_used_gb = data['max_percent_used_gb']
    max_avail_gb = data['max_avail_gb']
    num_views = data['num_views']
    num_channels = data['num_channels']
    num_det_rows = data['num_det_rows']
    num_indices = data['num_indices']

    # Set up for projections
    start_angle = 0.0
    end_angle = jnp.pi

    i = np.where(num_det_rows == nr)
    j = np.where(num_channels == nc)
    k = np.where(num_views == nv)

    for l, ni in enumerate(num_indices):
        angles = jnp.linspace(start_angle, end_angle, nv, endpoint=False)

        # Set up parallel beam
        sinogram_shape = (nv, nr, nc)
        parallel_model = mj.ParallelBeamModel(sinogram_shape, angles)

        parallel_model.pixel_batch_size_for_vmap = ni
        parallel_model.views_per_batch = nv
        parallel_model.main_device = jax.devices('cpu')[0]
        parallel_model.worker = jax.devices('gpu')[0]

        # Generate phantom for forward projection
        recon_shape = parallel_model.get_params('recon_shape')

        # Get a subset of the given size
        indices = np.arange(ni, dtype=int)
        voxel_values = np.ones((ni, recon_shape[2]))

        if eval_type_index == 0:
            print(
                'Initial forward projection for memory: nr={}, nc={}, nv={}, ni={}'.format(nr, nc, nv, ni))
            try:
                sinogram = parallel_model.sparse_forward_project(voxel_values, indices)
                m1 = mj.get_memory_stats()
                peak_mem_gb = m1[0]['peak_bytes_in_use'] / (1024 ** 3)
            except MemoryError as e:
                print('Out of memory')
                peak_mem_gb = 1000 * max_avail_gb
            mem_values[i, j, k, l] = peak_mem_gb
            print('Peak GB used = {}'.format(peak_mem_gb))

            print('Forward projection for speed')
            t0 = time.time()
            try:
                sinogram = parallel_model.sparse_forward_project(voxel_values, indices).block_until_ready()
                sinogram = parallel_model.sparse_forward_project(voxel_values, indices).block_until_ready()
                sinogram = parallel_model.sparse_forward_project(voxel_values, indices).block_until_ready()
            except MemoryError as e:
                print('Out of memory on pass 2')
            t1 = time.time()
            time_diff_secs = (t1 - t0) / 3
            time_values[i, j, k, l] = time_diff_secs
            print('Elapsed time = {}'.format(time_diff_secs))

        else:
            print('Initial back projection for memory: nv={}, nc={}, nr={}, ni={}'.format(nv, nc, nr, ni))
            sinogram = np.ones((nv, nr, nc))
            try:
                bp = parallel_model.sparse_back_project(sinogram, indices)
                m1 = mj.get_memory_stats()
                peak_mem_gb = m1[0]['peak_bytes_in_use'] / (1024 ** 3)
            except MemoryError as e:
                print('Out of memory')
                peak_mem_gb = 1000 * max_avail_gb

            mem_values[i, j, k, l] = peak_mem_gb
            print('Peak GB used = {}'.format(peak_mem_gb))

            print('Back projection for speed')
            t0 = time.time()
            try:
                bp = parallel_model.sparse_back_project(sinogram, indices).block_until_ready()
                bp = parallel_model.sparse_back_project(sinogram, indices).block_until_ready()
                bp = parallel_model.sparse_back_project(sinogram, indices).block_until_ready()
                t1 = time.time()
                time_diff_secs = (t1 - t0) / 3
            except:
                print('Out of memory on pass 2')
                t1 = time.time()
                time_diff_secs = 1000
            time_values[i, j, k, l] = time_diff_secs
            print('Elapsed time = {}'.format(time_diff_secs))
        max_percent_used_gb = max(max_percent_used_gb, 100 * peak_mem_gb / max_avail_gb)

    print('Max percentage GB used = {}%'.format(max_percent_used_gb))

    np.savez(filename, mem_values=mem_values, time_values=time_values, eval_type_index=np.array(eval_type_index),
             max_percent_used_gb=np.array(max_percent_used_gb), max_avail_gb=np.array(max_avail_gb),
             num_views=np.array(num_views), num_channels=np.array(num_channels),
             num_det_rows=np.array(num_det_rows), num_indices=np.array(num_indices))


if __name__ == '__main__':
    # filename, nv, nc, nr
    filename = sys.argv[1]
    nv = int(sys.argv[2])
    nc = int(sys.argv[3])
    nr = int(sys.argv[4])

    evaluate_over_indices(filename, nv, nc, nr)



import tempfile
import unittest
import numpy as np
import jax
import jax.numpy as jnp
import mbirjax as mj


class TestProjectors(unittest.TestCase):
    """
    Test the adjoint property of the forward and back projectors, both the full versions and the sparse voxel version.
    This means if x is an image, and y is a sinogram, then <y, Ax> = <Aty, x>.
    The code below verifies this for the full forward and back projectors and the versions that specify a
    subset of voxels in x.
    This code also verifies that first applying the full back projector and selecting the voxels to get (Aty)[ss]
    is the same as using the subset back projector with the specified set of voxels.
    """

    def setUp(self):
        """Set up before each test method."""
        # Choose the geometry type
        self.geometry_types = mj._utils._geometry_types_for_tests

        # Set parameters
        self.num_views = 64
        self.num_det_rows = 40
        self.num_det_channels = 128
        self.sharpness = 0.0

        # These can be adjusted to describe the geometry in the cone beam case.
        # np.Inf is an allowable value, in which case this is essentially parallel beam
        self.source_detector_dist = 4 * self.num_det_channels
        self.source_iso_dist = self.source_detector_dist

        # Initialize sinogram
        self.sinogram_shape = (self.num_views, self.num_det_rows, self.num_det_channels)
        self.angles = None
        self.translation_vector = None

    def tearDown(self):
        """Clean up after each test method."""
        pass

    def set_angles(self, geometry_type):
        if geometry_type == 'cone':
            detector_cone_angle = 2 * np.arctan2(self.num_det_channels / 2, self.source_detector_dist)
        else:
            detector_cone_angle = 0
        start_angle = -(np.pi + detector_cone_angle) * (1 / 2)
        end_angle = (np.pi + detector_cone_angle) * (1 / 2)
        self.angles = jnp.linspace(start_angle, end_angle, self.num_views, endpoint=False)

    def set_translation_vectors(self, geometry_type):
        if geometry_type == 'translation':
            self.translation_vectors = np.zeros((self.num_views, 3))
            self.translation_vectors[:, 0] = np.random.uniform(-10, 10, self.num_views)
            self.translation_vectors[:, 1] = 0.0
            self.translation_vectors[:, 2] = np.random.uniform(-10, 10, self.num_views)
            self.translation_vectors = jnp.array(self.translation_vectors)
        else:
            self.translation_vectors = None

    def get_model(self, geometry_type):
        if geometry_type == 'cone':
            ct_model = mj.ConeBeamModel(self.sinogram_shape, self.angles,
                                             source_detector_dist=self.source_detector_dist,
                                             source_iso_dist=self.source_iso_dist)
        elif geometry_type == 'parallel':
            ct_model = mj.ParallelBeamModel(self.sinogram_shape, self.angles)
        elif geometry_type == 'translation':
            ct_model = mj.TranslationModel(self.sinogram_shape, self.translation_vectors,
                                                source_detector_dist=self.source_detector_dist,
                                                source_iso_dist=self.source_iso_dist)
        else:
            raise ValueError('Invalid geometry type.  Expected cone or parallel, got {}'.format(geometry_type))

        return ct_model

    def test_all_adjoints(self):
        for geometry_type in self.geometry_types:
            with self.subTest(geometry_type=geometry_type):
                print("Testing adjoint with", geometry_type)
                self.verify_adjoint(geometry_type)

    def test_all_hessians(self):
        for geometry_type in self.geometry_types:
            with self.subTest(geometry_type=geometry_type):
                print("Testing Hessian with", geometry_type)
                self.verify_hessian(geometry_type)

    def test_save_load(self):
        for geometry_type in self.geometry_types:
            with self.subTest(geometry_type=geometry_type):
                print("Testing save/load with", geometry_type)
                self.verify_save_load(geometry_type)

    def verify_save_load(self, geometry_type):
        """
        Verify the adjoint property of the projectors:
        Choose a random phantom, x, and a random sinogram, y, and verify that <y, Ax> = <Aty, x>.
        """
        self.set_angles(geometry_type)
        self.set_translation_vectors(geometry_type)
        ct_model = self.get_model(geometry_type)

        # Generate phantom
        recon_shape = ct_model.get_params('recon_shape')
        num_recon_rows, num_recon_cols, num_recon_slices = recon_shape[:3]

        # Get the vector of indices
        indices = jnp.arange(num_recon_rows * num_recon_cols)

        # ##########################
        # Do a forward and back projection from a single pixel
        i, j = num_recon_rows // 4, num_recon_cols // 3
        x = jnp.zeros(recon_shape)
        x = x.at[i, j, :].set(1)
        voxel_values = x.reshape((-1, num_recon_slices))[indices]

        Ax = ct_model.sparse_forward_project(voxel_values, indices)
        Aty = ct_model.sparse_back_project(Ax, indices)
        Aty = ct_model.reshape_recon(Aty)

        # Save the model
        with tempfile.NamedTemporaryFile('w', suffix='.yaml') as file:
            filename = file.name
            ct_model.to_file(filename)

            # Load the model
            new_model = self.get_model(geometry_type)
            new_model = new_model.from_file(filename)

        # Compare parameters
        same_params = mj.ParameterHandler.compare_parameter_handlers(ct_model, new_model)
        assert same_params

        # Do a forward and back projection with loaded model
        Ax_new = new_model.sparse_forward_project(voxel_values, indices)
        Aty_new = new_model.sparse_back_project(Ax_new, indices)
        Aty_new = new_model.reshape_recon(Aty_new)

        # Compare to original
        assert(np.allclose(Aty, Aty_new, atol=1e-4))

    def test_view_batching(self):
        for geometry_type in self.geometry_types:
            with self.subTest(geometry_type=geometry_type):
                print("Testing view batching with", geometry_type)
                self.verify_view_batching(geometry_type)

    def verify_view_batching(self, geometry_type):
        self.set_angles(geometry_type)
        self.set_translation_vectors(geometry_type)
        ct_model = self.get_model(geometry_type)

        # Generate phantom
        recon_shape = ct_model.get_params('recon_shape')
        phantom = mj.gen_cube_phantom(recon_shape)

        # Generate indices of pixels and get the voxel cylinders
        full_indices = mj.gen_pixel_partition(recon_shape, num_subsets=1)[0]
        voxel_values = phantom.reshape((-1,) + recon_shape[2:])[full_indices]

        # Compute forward projection with all the views at once
        sinogram = ct_model.sparse_forward_project(voxel_values, full_indices)

        # Then compute the sinogram over multiple batches and reassemble them
        num_views = sinogram.shape[0]
        num_subsets = np.random.randint(2, 8)
        view_subsets = [jnp.arange(j, num_views, num_subsets) for j in range(num_subsets)]  # We don't use array_split because we want the entries to be interleaved for testing.
        sinogram_batched = [ct_model.sparse_forward_project(voxel_values, full_indices, view_indices=view_subsets[j]) for j in range(num_subsets)]
        sinogram_stitched = np.zeros_like(sinogram)
        for j in range(sinogram.shape[0]):
            sinogram_stitched[j] = sinogram_batched[j % num_subsets][j // num_subsets]

        forward_view_batch_test_result = np.allclose(sinogram, sinogram_stitched, atol=1e-5)
        self.assertTrue(forward_view_batch_test_result)

        # Then repeat for back projection
        back_projection = ct_model.sparse_back_project(sinogram, full_indices)
        back_projection_batched = [ct_model.sparse_back_project(sinogram, full_indices, view_indices=view_subsets[j]) for j in range(num_subsets)]
        back_projection_batched = np.stack(back_projection_batched, axis=0)
        back_projection_stitched = np.sum(back_projection_batched, axis=0)
        proj_diff = np.abs(back_projection_stitched - back_projection)
        # # The following is designed to highlight the bug associated with rounding in jax.
        # if np.sum(proj_diff > 1e-4) > 10:
        #     print('Num above threshold = {}, max diff = {}'.format(np.sum(proj_diff > 1e-4), np.amax(proj_diff)))
        #     row_index0, col_index0 = jnp.unravel_index(full_indices, recon_shape[:2])
        #     recon0 = jnp.zeros(recon_shape)
        #     recon0 = recon0.at[row_index0, col_index0].set(back_projection)
        #     recon1 = jnp.zeros(recon_shape)
        #     recon1 = recon1.at[row_index0, col_index0].set(back_projection_stitched)
        #     title = 'Standard backprojection (left) and \nabs diff with back projection via multiple view subsets (right)'
        #     title += '\nDifferences are due to inconsistent choices of rounding in jax.  See experiments/bugs_and_artifacts'
        #     mj.slice_viewer(recon0, recon1-recon0, slice_axis=2, vmax=0.2, title=title)
        back_view_batch_test_result = np.sum(proj_diff > 1e-4) < 1000 and np.amax(proj_diff) < 0.2
        self.assertTrue(back_view_batch_test_result)

    def verify_adjoint(self, geometry_type):
        """
        Verify the adjoint property of the projectors:
        Choose a random phantom, x, and a random sinogram, y, and verify that <y, Ax> = <Aty, x>.
        """
        self.set_angles(geometry_type)
        self.set_translation_vectors(geometry_type)
        ct_model = self.get_model(geometry_type)

        # Initialize a random key
        seed_value = np.random.randint(1000000)
        key = jax.random.PRNGKey(seed_value)

        # Generate phantom
        recon_shape = ct_model.get_params('recon_shape')
        num_recon_rows, num_recon_cols, num_recon_slices = recon_shape[:3]
        phantom = mj.gen_cube_phantom(recon_shape)

        # Generate indices of pixels
        num_subsets = 1
        full_indices = mj.gen_pixel_partition(recon_shape, num_subsets)

        # Generate sinogram data
        voxel_values = phantom.reshape((-1,) + recon_shape[2:])[full_indices]

        # Compute forward projection
        sinogram = ct_model.sparse_forward_project(voxel_values[0], full_indices[0])

        # Get the vector of indices
        indices = jnp.arange(num_recon_rows * num_recon_cols)
        num_trials = 3
        indices = jnp.mod(np.arange(num_trials, dtype=int).reshape((-1, 1)) + indices.reshape((1, -1)),
                          num_recon_rows * num_recon_cols)

        # Convert to jax arrays
        sinogram = jnp.array(sinogram)
        indices = jnp.array(indices)

        # Run once to finish compiling and get backprojection shape
        bp = ct_model.sparse_back_project(sinogram, indices[0])

        # ##########################
        # Test the adjoint property
        # Get a random 3D phantom to test the adjoint property
        key, subkey = jax.random.split(key)
        x = jax.random.uniform(subkey, shape=bp.shape)
        key, subkey = jax.random.split(key)
        y = jax.random.uniform(subkey, shape=sinogram.shape)

        # Do a forward projection, then a backprojection
        voxel_values = x.reshape((-1, num_recon_slices))[indices[0]]
        Ax = ct_model.sparse_forward_project(voxel_values, indices[0])
        Aty = ct_model.sparse_back_project(y, indices[0])

        # Calculate <Aty, x> and <y, Ax>
        Aty_x = jnp.sum(Aty * x)
        y_Ax = jnp.sum(y * Ax)

        # Determine if property holds
        adjoint_test_result = np.allclose(Aty_x, y_Ax, rtol=1e-4)
        print("maximum difference = ", np.max(Aty_x - y_Ax))
        print("minimum difference = ", np.min(Aty_x - y_Ax))
        self.assertTrue(adjoint_test_result)

    def verify_hessian(self, geometry_type):
        """
        Verify the hessian property of the back projector:
        Choose a random pixel, set it to epsilon, apply A^T A and compare to the value from compute_hessian_diagaonal.
        """
        self.set_angles(geometry_type)
        self.set_translation_vectors(geometry_type)
        ct_model = self.get_model(geometry_type)

        # ## Test the hessian against a finite difference approximation ## #
        hessian = ct_model.compute_hessian_diagonal()

        # Initialize a random key
        seed_value = np.random.randint(1000000)
        key = jax.random.PRNGKey(seed_value)

        recon_shape = ct_model.get_params('recon_shape')
        num_recon_rows, num_recon_cols, num_recon_slices = recon_shape[:3]
        x = jnp.zeros(recon_shape)
        key, subkey = jax.random.split(key)
        i, j = jax.random.randint(subkey, shape=(2,), minval=0, maxval=num_recon_rows)
        key, subkey = jax.random.split(key)
        k = jax.random.randint(subkey, shape=(), minval=0, maxval=num_recon_slices)

        # Get the vector of indices
        indices = jnp.arange(num_recon_rows * num_recon_cols)
        # num_trials = 3
        # indices = jnp.mod(np.arange(num_trials, dtype=int).reshape((-1, 1)) + indices.reshape((1, -1)),
        #                   num_recon_rows * num_recon_cols)

        eps = 0.01
        x = x.at[i, j, k].set(eps)
        voxel_values = x.reshape((-1, num_recon_slices))[indices]
        Ax = ct_model.sparse_forward_project(voxel_values, indices)
        AtAx = ct_model.sparse_back_project(Ax, indices).reshape(x.shape)
        finite_diff_hessian = AtAx[i, j, k] / eps

        # Determine if property holds
        hessian_test_result = jnp.allclose(hessian.reshape(x.shape)[i, j, k], finite_diff_hessian)
        self.assertTrue(hessian_test_result)


if __name__ == '__main__':
    unittest.main()



import numpy as np
import jax
import jax.numpy as jnp
import mbirjax as mj
import unittest


class TestProx(unittest.TestCase):
    """
    Test the proximal map function for the special case when the input and initial condition are both the ground truth.
    """

    def setUp(self):
        """Set up before each test method."""
        np.random.seed(0)  # Set a seed to avoid variations due to partition creation.

        # Set parameters
        self.num_views = 32
        self.num_det_rows = 40
        self.num_det_channels = 128
        self.sharpness = 0.0

        # Initialize sinogram
        self.sinogram_shape = (self.num_views, self.num_det_rows, self.num_det_channels)
        self.angles = None

    def tearDown(self):
        """Clean up after each test method."""
        pass

    def test_prox(self):
        """
        Verify that the proximal map function works for a simple corner case.
        """
        start_angle = 0.0
        end_angle = np.pi

        self.angles = jnp.linspace(start_angle, end_angle, self.num_views, endpoint=False)
        ct_model = mj.ParallelBeamModel(self.sinogram_shape, self.angles)

        # Generate 3D Shepp Logan phantom
        print('  Creating phantom')
        phantom_shape = ct_model.get_params('recon_shape')
        phantom = mj.generate_3d_shepp_logan_low_dynamic_range(phantom_shape)

        # Generate synthetic sinogram data
        print('  Creating sinogram')
        sinogram = ct_model.forward_project(phantom)

        # Set reconstruction parameter values
        ct_model.set_params(verbose=0)

        # ##########################
        # Evaluating the proximal map
        print('  Starting proximal map')
        prox_input = phantom + 0.1 * (phantom > 0)

        ct_model.set_params(sharpness=6, snr_db=60)
        recon, recon_dict = ct_model.recon(sinogram, max_iterations=20, first_iteration=0,
                                           stop_threshold_change_pct=0.01)

        # Do a prox map with small sigma_prox to return very nearly the prox_input
        prox_recon0, prox_recon_dict0 = ct_model.prox_map(prox_input, sinogram, sigma_prox=1e-6,
                                                          max_iterations=20,
                                                          first_iteration=0, init_recon=prox_input)

        # Do a prox map with large sigma_prox to return very nearly the same as a recon with small prior
        prox_recon1, prox_recon_dict1 = ct_model.prox_map(prox_input, sinogram, sigma_prox=1e6,
                                                          max_iterations=20,
                                                          first_iteration=0, init_recon=recon)

        small_sigma_nrmse = np.linalg.norm(prox_recon0 - prox_input) / np.linalg.norm(prox_input)
        large_sigma_nrmse = np.linalg.norm(prox_recon1 - recon) / np.linalg.norm(recon)
        tolerance = 2e-5
        self.assertTrue(small_sigma_nrmse < tolerance)
        tolerance = 3e-4
        self.assertTrue(large_sigma_nrmse < tolerance)


if __name__ == '__main__':
    unittest.main()



import numpy as np
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import sys


def display_results(filename):

    # Load the existing data
    data = np.load(filename, allow_pickle=True)
    mem_values = data['mem_values']
    time_values = data['time_values']
    eval_type_index = data['eval_type_index']
    max_percent_used_gb = data['max_percent_used_gb']
    max_avail_gb = data['max_avail_gb']
    num_views = data['num_views']
    num_channels = data['num_channels']
    num_det_rows = data['num_det_rows']
    num_indices = data['num_indices']

    # Set up the info for plotting
    outer_labels = ['# det rows', '# det channels']
    outer_values = [num_det_rows, num_channels]
    inner_labels = ['# views', '# indices']
    inner_values = [num_views, num_indices]

    vmin_mem = -2
    vmax_mem = 6
    vmin_time = -12
    vmax_time = 5

    mem0 = mem_values.flatten()[0]
    mem_values /= mem0

    time_values = normalize_time_values(time_values, outer_labels, outer_values, inner_labels, inner_values)
    time0 = time_values.flatten()[0]
    time_values /= time0

    plt.ion()

    eval_types = ['Forward_projection', 'Backward_projection']
    super_title_mem = eval_types[eval_type_index]
    super_title_mem += '\nlog2(GB / GB_baseline) for one batch as a function of #rows, #channels, #views, and #indices'
    super_title_mem += '\nGB_baseline = {:.2f}GB'.format(mem0)
    # super_title_mem += ':  Avail GB = {:.1f}, Max percent mem used = {:.2f}'.format(max_avail_gb, max_percent_used_gb)
    super_title_time = eval_types[eval_type_index]
    super_title_time += '\nlog2(seconds elapsed / secs_baseline) for full set of batches as a function of #rows, #channels, #views, and #indices'
    super_title_time += '\nsecs_baseline elapsed = {:.3g} secs'.format(time0)

    create_tiled_heatmap(np.log2(mem_values), outer_labels, outer_values, inner_labels, inner_values,
                         super_title=super_title_mem, vmin=vmin_mem, vmax=vmax_mem)
    create_tiled_heatmap(np.log2(time_values), outer_labels, outer_values, inner_labels, inner_values,
                         super_title=super_title_time, vmin=vmin_time, vmax=vmax_time)


def normalize_time_values(time_values, outer_labels, outer_values, inner_labels, inner_values):
    """
    Each entry in time_values gives the time needed for a batch of views and indices for a given umber of detector rows
    and channels.  Assuming a recon size of num_channels**2 * num_rows and a sinogram with num_channels views, this
    function multiplies each entry by the number of reps needed to project the full recon and sino.
    Args:
        time_values:
        outer_labels:
        outer_values:
        inner_labels:
        inner_values:

    Returns:

    """
    row_index = 0
    channel_index = 1
    assert(outer_labels[row_index] == '# det rows' and outer_labels[channel_index] == '# det channels')
    view_index = 0
    indices_index = 1
    assert(inner_labels[view_index] == '# views' and inner_labels[indices_index] == '# indices')
    for i, num_row in enumerate(outer_values[row_index]):
        for j, num_channels in enumerate(outer_values[channel_index]):
            total_indices = num_channels * num_channels * 0.8  # Approx # pixels in the region of reconstruction
            total_views = num_channels
            for k, num_views in enumerate(inner_values[view_index]):
                for n, num_indices in enumerate(inner_values[indices_index]):
                    num_reps = (total_views / num_views) * (total_indices / num_indices)
                    time_values[i, j, k, n] *= num_reps
    return time_values


def create_tiled_heatmap(plot_values, outer_labels, outer_values, inner_labels, inner_values,
                         super_title=None, vmin=None, vmax=None):
    """
    Creates a nested grid to display the 4D data in plot_values, assuming a shape on the order of 3x3x3x3, where
    each axis of plot_values is assumed to correspond to an independent variable as described in outer_labels and
    inner_labels.  The first 2 axes of plot values determine the location within an outer grid, and the next 2 axes
    determine the location within each inner grid.  For each [i,j], the values in plot_values[i, j] are plotted using
    a colormap along with the text of the value in each grid square.

    plot_values is assumed to be obtained by evaluating
    a function f at points determined by inner_values and outer_values as follows:
    plot_values[i, j, k, l] = f[outer_values[0][i], outer_values[1][j], inner_values[0][k], inner_values[1][l]]

    Args:
        plot_values (4D numpy array):  Values to display
        outer_labels (length 2 list of strings): The labels describing axes 0 and 1 of plot_values (e.g., #views, #channels)
        outer_values (length 2 list of lists of int):  outer_values[0] (or [1]) contains the values of the independent variable for axis 0 (or 1)
        inner_labels (length 2 list of strings): The labels describing axes 2 and 3 of plot_values (e.g., #slices, #indices)
        inner_values (length 2 list of lists int): inner_values[0] (or [1]) contains the values of the independent variable for axis 2 (or 3)
        sup_title (string, optional): Overall title for the display
        vmin (float, optional): lower clipping value for colorbar display
        vmax (float, optional): upper clipping value for colorbar display

    Returns:
        Nothing

    # Example:
    plot_values = 15*np.random.rand(3, 3, 3, 3)
    outer_labels = ['# views', '# det channels']
    outer_values = [[128, 256, 512], [128, 256, 512]]
    inner_labels = ['# det rows', '# indices']
    inner_values = [[128, 256, 512], [129, 256, 512]]
    super_title = 'Peak log2(GB) as a function of views, channels, rows, and #indices'
    vmin = 0.0
    vmax = 10.0
    create_tiled_heatmap(np.log2(plot_values), outer_labels, outer_values, inner_labels, inner_values,
                         super_title=super_title, vmin=vmin, vmax=vmax)
    """
    if vmin is None:
        vmin = np.amin(plot_values)
    if vmax is None:
        vmax = np.amax(plot_values)
    # Create a grid for the subplots with some space between them
    nrows = plot_values.shape[0] + 1
    ncols = plot_values.shape[1] + 1
    width_ratios = [1, ] + (ncols - 1) * [3, ]
    height_ratios = (nrows - 1) * [3, ] + [1, ]

    gs = gridspec.GridSpec(nrows=nrows, ncols=ncols, width_ratios=width_ratios, height_ratios=height_ratios,
                           hspace=0.0, wspace=0.5)

    # Convert all displayed items to strings
    inner_labels = [str(label) for label in inner_labels]
    row_names = [str(row_name) for row_name in inner_values[0]]
    col_names = [str(col_name) for col_name in inner_values[1]]
    inner_values = [row_names, col_names]

    outer_labels = [str(sup_label) for sup_label in outer_labels]
    sup_row_names = [str(row_name) for row_name in outer_values[0]]
    sup_col_names = [str(col_name) for col_name in outer_values[1]]
    outer_values= [sup_row_names, sup_col_names]

    # Set up the overall figure
    fig, axs = plt.subplots(figsize=(5*(ncols-1), 5*(nrows-1)))
    cax = None
    fig.get_axes()[0].axis('off')
    # Create a subplot for each 2D array in 'plot_values'
    for i in range(nrows):
        for j in range(ncols):
            ax = fig.add_subplot(gs[i, j])

            # Display the row outer labels
            if j == 0:
                ax.axis('off')
                if i < nrows-1:
                    ax.text(0.5, 0.5, outer_labels[0] + '\n' + outer_values[0][i], horizontalalignment='center', verticalalignment='center',
                            transform=ax.transAxes)
            # Display the column outer labels
            elif i == nrows-1:
                ax.axis('off')
                ax.text(0.5, 0.5, outer_labels[1] + '\n' + outer_values[1][j-1], horizontalalignment='center', verticalalignment='center',
                        transform=ax.transAxes)
            else:
                cur_values = plot_values[i, j-1]  # Offset by one since the leftmost subplot was used for a label
                # Create the heatmap
                cax = ax.matshow(cur_values, cmap='viridis', vmin=vmin, vmax=vmax)

                # Set the inner tick labels for rows and columns
                ax.set_yticks(np.arange(len(inner_values[0])), inner_values[0])
                ax.set_xticks(np.arange(len(inner_values[1])), inner_values[1])
                ax.xaxis.set_ticks_position('bottom')

                # Set the inner axes labels
                ax.set_ylabel(inner_labels[0])
                ax.set_xlabel(inner_labels[1])

                # Display the plot_values in the squares
                for m in range(cur_values.shape[0]):
                    for n in range(cur_values.shape[1]):
                        ax.text(n, m, f'{cur_values[m, n]:.1f}', ha='center', va='center', color='w')

    # Add a colorbar on the right and the super title
    fig.colorbar(cax, ax=fig.get_axes())
    if super_title is not None:
        plt.suptitle(super_title)

    plt.show()


if __name__ == '__main__':
    # filename, nv, nc, nr
    filename = sys.argv[1]

    display_results(filename)
    plt.pause(5)
    input('Press return to exit')



import numpy as np
import time
import jax.numpy as jnp
import matplotlib.pyplot as plt
import mbirjax as mj


def display_slices_for_abstract( recon1, recon2, recon3, labels) :
    # Set global font size
    plt.rcParams.update({'font.size': 15})  # Adjust font size here

    vmin = 0.0
    vmax = phantom.max()
    slice_index = recon1.shape[2] // 2

    fig, ax = plt.subplots(nrows=1, ncols=3, figsize=(15, 5))

    a0 = ax[0].imshow(recon1[:, :, slice_index], vmin=vmin, vmax=vmax, cmap='gray')
    #plt.colorbar(a0, ax=ax[0])
    ax[0].set_title(labels[0])

    a0 = ax[1].imshow(recon2[:, :, slice_index], vmin=vmin, vmax=vmax, cmap='gray')
    #plt.colorbar(a0, ax=ax[1])
    ax[1].set_title(labels[1])

    a2 = ax[2].imshow(recon3[:, :, slice_index], vmin=vmin, vmax=vmax, cmap='gray')
    #plt.colorbar(a2, ax=ax[2])
    ax[2].set_title(labels[2])

    plt.show()


if __name__ == "__main__":
    """
    This is a script to develop, debug, and tune the vcd reconstruction with a parallel beam projector
    """
    # Set parameters
    num_views = 256
    num_det_rows = 10
    num_det_channels = 256
    start_angle = 0
    end_angle = np.pi
    sharpness = 0.0

    # Initialize sinogram
    sinogram = jnp.zeros((num_views, num_det_rows, num_det_channels))
    angles = jnp.linspace(start_angle, np.pi, num_views, endpoint=False)

    # Set up parallel beam model
    parallel_model = mj.ParallelBeamModel(sinogram.shape, angles)

    # Generate 3D Shepp Logan phantom
    phantom_shape = parallel_model.get_params('recon_shape')
    phantom = mj.generate_3d_shepp_logan_low_dynamic_range(phantom_shape)

    # Generate synthetic sinogram data
    sinogram = parallel_model.forward_project(phantom)

    # Generate weights array
    weights = mj.gen_weights(sinogram / sinogram.max(), weight_type='transmission_root')

    # Set reconstruction parameter values
    parallel_model.set_params(sharpness=sharpness, verbose=1)
    parallel_model.set_params(partition_sequence=[0, 1, 2, 3, 1, 2, 3, 2, 3, 3, 0, 1, 2, 3, 1, 2, 3, 2, 3, 3])
    granularity = np.array([1, 8, 64, 256])

    # Print out model parameters
    parallel_model.print_params()
    init_recon = np.zeros(phantom.shape)
    # ##########################
    # Perform VCD reconstruction
    max_iterations = 10
    recon_vcd, recon_dict_vcd = parallel_model.recon(sinogram, init_recon=init_recon, weights=weights, max_iterations=max_iterations,
                                                       compute_prior_loss=True)
    recon_params_vcd = recon_dict_vcd['recon_params']
    fm_rmse_vcd = recon_params_vcd['fm_rmse']
    prior_loss_vcd = recon_params_vcd['prior_loss']
    default_partition_sequence = parallel_model.get_params('partition_sequence')
    partition_sequence = mj.gen_partition_sequence(default_partition_sequence, max_iterations=max_iterations)
    granularity_sequence_vcd = granularity[partition_sequence]

    # Perform GD reconstruction
    partition_sequence = [0, ]
    parallel_model.set_params(partition_sequence=partition_sequence)
    recon_prev_default, recon_dict_prev_default = parallel_model.recon(sinogram, init_recon=init_recon, weights=weights, max_iterations=max_iterations,
                                                                         compute_prior_loss=True)
    recon_params_prev_default = recon_dict_prev_default['recon_params']
    fm_rmse_prev_default = recon_params_prev_default['fm_rmse']
    prior_loss_prev_default = recon_params_prev_default['prior_loss']
    partition_sequence = mj.gen_partition_sequence(partition_sequence=partition_sequence, max_iterations=max_iterations)
    granularity_sequence_prev_default = granularity[partition_sequence]

    # Perform CD reconstruction
    partition_sequence = [3, ]
    parallel_model.set_params(partition_sequence=partition_sequence)
    recon_cd, recon_dict_cd = parallel_model.recon(sinogram, init_recon=init_recon, weights=weights, max_iterations=max_iterations,
                                                     compute_prior_loss=True)
    recon_params_cd = recon_dict_cd['recon_params']
    fm_rmse_cd = recon_params_cd['fm_rmse']
    prior_loss_cd = recon_params_cd['prior_loss']
    partition_sequence = mj.gen_partition_sequence(partition_sequence=partition_sequence, max_iterations=max_iterations)
    granularity_sequence_cd = granularity[partition_sequence]
    # ##########################

    # Display reconstructions
    labels = ['Gradient Descent', 'Vectorized Coordinate Descent', 'Coordinate Descent']
    display_slices_for_abstract(recon_prev_default, recon_vcd, recon_cd, labels)

    # Display granularity plots:
    granularity_sequences = [granularity_sequence_prev_default, granularity_sequence_vcd, granularity_sequence_cd]
    fm_losses = [fm_rmse_prev_default, fm_rmse_vcd, fm_rmse_cd]
    prior_losses = [prior_loss_prev_default, prior_loss_vcd, prior_loss_cd]
    labels = ['Gradient Descent', 'Vectorized Coordinate Descent', 'Coordinate Descent']
    mj.plot_granularity_and_loss(granularity_sequences, fm_losses, prior_losses, labels, granularity_ylim=(0, 256), loss_ylim=(0.1, 15))

    # Generate sequence of partition images for Figure 1
    recon_shape = (32, 32, 1)
    partitions_fig = mj.gen_set_of_pixel_partitions(recon_shape=recon_shape, granularity=[1, 4, 16, 64, 256])

    # Plot the set of partitions
    mj.debug_plot_partitions(partitions=partitions_fig, recon_shape=recon_shape)



#!/usr/bin/env python
import argparse
import csv
import itertools
import subprocess
import sys
import numpy as np
import time
import os
import mbirjax as mj

target_gb = 40

max_gb = 79
os.environ['XLA_PYTHON_CLIENT_MEM_FRACTION'] = str(target_gb / max_gb)

"""
This is a Python program to run a grid search over parameters and record the peak GPU memory usage from a JAX computation.
The code to run is either a direct recon, or a projection, or a recon with specified use of the GPU. 
The parameters to be scanned are 

    v: number of views
    r: number of rows
    c: number channels
    b: batch size
    
Additionally, the memory fraction for jax should be varied since memory use depends on available memory.  This 
can be done manually using e.g.

target_gb = 10
max_gb = 79
os.environ['XLA_PYTHON_CLIENT_MEM_FRACTION'] = str(target_gb / max_gb)

How This Code Works

For each fixed combination of (v, r, c), the program starts a new process to evaluate the function and record the 
memory and time in a .csv file.  Within this process, the batch size is increased so that memory use is monotonic, 
with termination in case of OOM.  

    Worker Process Mode:

        When run with the --worker flag and with parameters --v, --r, and --c, the script enters worker_main().
        For each value of b (using the same grid as for the other parameters), it runs the JAX function and obtains the peak memory usage via device.memory_stats().
        The worker prints one CSV row (with comma-separated values for v, r, c, b, and peak_memory) per iteration. Any errors are printed to stderr.

    Main Process Mode:

        When no --worker flag is provided, the script runs in main mode.
        It opens (or creates) a CSV file (memory_stats.csv) and writes a header row.
        It then loops over all combinations of v, r, and c (using itertools.product), spawns a worker process for each combination (using subprocess.run), and captures the worker’s stdout.
        The main process parses each CSV row output by the worker and writes it to the CSV file.
        If any worker prints error messages to stderr, these are forwarded to the main process’s stderr.
"""

# Define the set of parameter values to test.
PARAM_VALUES = [50, 100, 250, 500, 800, 1250]
BATCH_SIZES = [100, 250, 500, 800]
OUTPUT_SYMBOL = "#"


def worker_main(v, r, c):
    """
    Worker process:
    For a fixed combination of v, r, c, loop over b values,
    run the JAX computation, and print a CSV row for each b.
    """
    import jax
    import jax.numpy as jnp

    # Define a JAX function that depends on the parameters.
    @jax.jit
    def sample_function(v, r, c, b, x):
        A = jnp.ones((v, r))
        B = jnp.ones((c, b))
        # For a valid dot product, x is chosen with shape (r, c)
        result = jnp.dot(A, jnp.dot(x, B))
        return result

    def run_experiment(v, r, c, b):

        # Create an input array x with shape (v, r, c)
        sino = np.zeros((v, r, c))
        sino[:, r//3:r-r//3, c//3:c-c//3] = 1
        angles = np.linspace(0, 2*np.pi, v, endpoint=False)
        source_detector_dist = 4 * c
        source_iso_dist = source_detector_dist

        ct_model = mj.ConeBeamModel(sino.shape, angles,
                                         source_detector_dist=source_detector_dist,
                                         source_iso_dist=source_iso_dist)
        ct_model.set_params(use_gpu='sinograms', verbose=0)
        ct_model.view_batch_size_for_vmap = b
        weights = mj.gen_weights(sino, weight_type='unweighted')

        # Run the computation and block until complete.
        time0 = time.time()
        recon, recon_dict = ct_model.recon(sino, weights=weights, max_iterations=1)
        _ = recon.block_until_ready()
        elapsed_time0 = time.time() - time0

        # time0 = time.time()
        # recon, recon_dict = ct_model.recon(sino, weights=weights, max_iterations=1)
        # _ = recon.block_until_ready()
        # elapsed_time1 = time.time() - time0
        elapsed_time1 = elapsed_time0

        # Retrieve memory stats.
        stats = mj.get_memory_stats(print_results=False)
        peak_memory = stats[0]['peak_bytes_in_use'] / (1024 ** 3)
        avail_memory = stats[0]['bytes_limit'] / (1024 ** 3)
        return peak_memory, avail_memory, elapsed_time0, elapsed_time1

    # Loop over b values
    summary_output = ""
    num_output_lines = 0
    for b in BATCH_SIZES:
        try:
            peak, avail, elapsed0, elapsed1 = run_experiment(v, r, c, b)
            # Print CSV row: v, r, c, b, peak_memory, elapsed_time.
            # These values will be captured by the main process.
            summary_output += OUTPUT_SYMBOL + f",{v},{r},{c},{b},{peak:.3f},{avail:.3f},{elapsed0:.1f},{elapsed1:.1f}\n"
            num_output_lines += 1
        except Exception as e:
            # If an error occurs, print to stderr.
            print(f"{v},{r},{c},{b},Error: {e}", file=sys.stderr)

    print(summary_output)


def main():
    """
    Main process:
    Iterate over all combinations of (v, r, c), spawn a worker process
    for each, and write the output to a CSV file.
    """
    # Open the output CSV file and write the header.
    output_file_name = "memory_stats.csv"
    file_exists = os.path.isfile(output_file_name)
    with open(output_file_name, "a", newline="") as csvfile:
        writer = csv.writer(csvfile)
        if not file_exists:
            writer.writerow(["num views", "num rows", "num channels", "batch size", "peak memory (GB)", "avail memory (GB)", "elapsed time0 (sec)", "elapsed time1 (sec)"])
        # Iterate over all combinations of v, r, c.
        for v, r, c in itertools.product(PARAM_VALUES, repeat=3):
            print('\nStarting multiple batches with v={}, r={}, c={}'.format(v, r, c))
            print('Batch sizes = {}'.format(BATCH_SIZES))
            # Spawn a worker process for the given (v, r, c) combination.
            result = subprocess.run(
                [sys.executable, __file__, "--worker", "--v", str(v), "--r", str(r), "--c", str(c)],
                capture_output=True,
                text=True
            )
            # Process the worker's stdout which contains CSV rows.
            for line in result.stdout.strip().splitlines():
                # Expect each line to have the format: v,r,c,b,peak_memory
                row = line.strip().split(",")
                if row[0] == OUTPUT_SYMBOL:
                    writer.writerow(row[1:])
            if result.stderr:
                print(f"Error in worker for v={v}, r={r}, c={c}:", result.stderr, file=sys.stderr)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="JAX GPU memory usage experiment.")
    parser.add_argument("--worker", action="store_true", help="Run as worker process")
    parser.add_argument("--v", type=int, help="Parameter v (required for worker)")
    parser.add_argument("--r", type=int, help="Parameter r (required for worker)")
    parser.add_argument("--c", type=int, help="Parameter c (required for worker)")
    args = parser.parse_args()

    if args.worker:
        # Worker mode: require all three parameters.
        if args.v is None or args.r is None or args.c is None:
            raise ValueError("Worker process requires --v, --r, and --c arguments.")
        worker_main(args.v, args.r, args.c)
    else:
        # Main mode: spawn worker processes and aggregate CSV output.
        main()



from collections import namedtuple
import jax
import jax.numpy as jnp
import mbirjax
from functools import partial


class Projectors:

    def __init__(self, tomography_model):

        self.tomography_model = tomography_model
        self.sparse_forward_project, self.sparse_back_project = None, None
        self.forward_project_pixel_batch, self.back_project_view_batch = None, None
        self.create_projectors(tomography_model)

    def create_projectors(self, tomography_model):
        """
        Compute the forward and back projectors for this geometry and current view parameters

        Args:
            tomography_model (mbirjax.TomographyModel): An instance describing the current geometry and implementing the following 2 functions:

                * forward_project_pixel_batch_to_one_view (callable): jit-compilable function implementing :meth:`TomographyModel.forward_project_pixel_batch_to_one_view`
                * back_project_one_view_to_pixel_batch (callable): jit-compilable function implementing :meth:`TomographyModel.back_project_one_view_to_pixel_batch`

        Returns:
            Nothing, but the class variables `sparse_forward_project` and `sparse_back_project` are set to callable
            functions.  These are used to implement the following methods:

            * `sparse_forward_project`: :meth:`TomographyModel.sparse_forward_project`
            * `sparse_back_project`: :meth:`TomographyModel.sparse_back_project`

        Note:
            The returned functions will be jit compiled each time they are called with a new shape of input.  If
            called multiple times with the same shape of input, then the cached version will be used, which will
            give reduced execution time relative to the initial call.

            This method requires geometry-specific implementations of
            :meth:`TomographyModel.forward_project_pixel_batch_to_one_view` and
            :meth:`TomographyModel.back_project_one_view_to_pixel_batch`.

        """
        forward_project_pixel_batch_to_one_view = tomography_model.forward_project_pixel_batch_to_one_view
        back_project_one_view_to_pixel_batch = tomography_model.back_project_one_view_to_pixel_batch

        geometry_params = self.tomography_model.get_geometry_parameters()
        sinogram_shape, recon_shape = self.tomography_model.get_params(['sinogram_shape', 'recon_shape'])

        # Combine the needed parameters into a named tuple for named access compatible with jit
        projector_param_names = ['sinogram_shape', 'recon_shape', 'geometry_params']
        projector_param_values = (sinogram_shape, recon_shape, geometry_params)
        ProjectorParams = namedtuple('ProjectorParams', projector_param_names)
        projector_params = ProjectorParams(*tuple(projector_param_values))

        view_params_name = self.tomography_model.get_params('view_params_name')
        view_params_array = self.tomography_model.get_params(view_params_name)
        pixel_batch_size = self.tomography_model.pixel_batch_size_for_vmap
        view_batch_size = self.tomography_model.view_batch_size_for_vmap

        def sparse_forward_project_fcn(voxel_values, pixel_indices, view_indices=()):
            """
            Compute the sinogram obtained by forward projecting the specified voxels. The voxel sare determined
            using 2D indices into a flattened array of shape (num_recon_rows, num_recon_cols),
            and for each such 2D index, the voxels in all slices at that location are projected.

            This function batches over pixels, applies sparse_forward_project_pixel_batch to each batch,
            then adds the results to get the sinogram.

            Args:
                voxel_values (ndarray or jax array): 2D array of shape (len(pixel_indices), num_slices) of voxel values
                pixel_indices (ndarray or jax array): 1D array of indices into a flattened array of shape (num_rows, num_cols)
                view_indices (ndarray or jax array, optional): 1D array of indices into the view parameters array.
                    If None, then all views are used.

            Returns:
                3D array of shape (num_views, num_det_rows, num_det_cols), where num_views is len(view_indices) if view_indices is not None
            """

            voxel_and_indices = (voxel_values, pixel_indices)  # Apply ensure_tuple

            def forward_project_pixel_batch_wrapper(local_values, local_pix_indices):
                return sparse_forward_project_pixel_batch(local_values, local_pix_indices, view_indices,
                                                          view_batch_size)

            summed_output = sum_function_in_batches(forward_project_pixel_batch_wrapper, voxel_and_indices,
                                                    pixel_batch_size)
            return summed_output

        @partial(jax.jit, static_argnames='views_per_batch')
        def sparse_forward_project_pixel_batch(voxel_values, pixel_indices, view_indices=(), views_per_batch=None):
            """
            Compute the sinogram obtained by forward projecting the specified batch of voxels. The voxels
            are determined using 2D indices into a flattened array of shape (num_rows, num_cols),
            and for each such 2D index, the voxels in all slices at that location are projected.

            This function creates batches of views and collects the results to form the full sinogram.

            Args:
                voxel_values: 2D array of shape (len(pixel_indices), num_slices) of voxel values
                pixel_indices: 1D array of indices into a flattened array of shape (num_rows, num_cols)
                view_indices (ndarray or jax array, optional): 1D array of indices into the view parameters array.
                    If None, then all views are used.
                views_per_batch (int or None, optional): Maximum number of views to project per batch.

            Returns:
                3D array of shape (num_views, num_det_rows, num_det_cols)
            """
            cur_view_params_array = view_params_array
            if len(view_indices) > 0:
                cur_view_params_array = view_params_array[view_indices]

            def forward_project_single_view(single_view_params):
                # Use closure to define a mappable function that operates on a single view with the given voxel values.
                return forward_project_pixel_batch_to_one_view(voxel_values, pixel_indices,
                                                               single_view_params, projector_params)

            def forward_project_view_batch(view_params_batch):
                # Map the single view function over a batch of views.
                # To parallelize over views, we can use jax.vmap here instead of jax.lax.map, but this may use extra
                # memory since the voxel values are required for each view.

                sino_view_batch = jax.vmap(forward_project_single_view)(view_params_batch)

                return sino_view_batch

            sinogram = concatenate_function_in_batches(forward_project_view_batch, cur_view_params_array, views_per_batch)

            return sinogram

        def sparse_back_project_fcn(sinogram, pixel_indices, coeff_power=1, view_indices=()):
            """
            Compute the voxel values obtained by back projecting the sinogram to the specified voxels. The voxels
            are determined using 2D indices into a flattened array of shape (num_recon_rows, num_recon_cols),
            and for each such 2D index, the voxels in all slices at that location are projected.

            This function batches over views, applies sparse_back_project_view_batch to each batch,
            then adds the results to get the voxel values.

            Args:
                sinogram (ndarray or jax array): 3D array of shape (cur_num_views, num_det_rows, num_det_cols), where
                    cur_num_views is recon_shape[0] if view_indices is () and len(view_indices)
                    otherwise, in which case the views in sinogram should match those indicated by view_indices.
                pixel_indices (ndarray or jax array): 1D array of indices into a flattened array of shape
                (num_recon_rows, num_recon_cols)
                coeff_power (int): backproject using the coefficients of (A_ij ** coeff_power).
                    Normally 1, but should be 2 when computing Hessian diagonal.
                view_indices (ndarray or jax array, optional): 1D array of indices into the view parameters array.
                    If None, then all views are used.

            Returns:
                2D array of shape (num_pixels, num_recon_slices)
            """

            cur_view_params_array = view_params_array
            if len(view_indices) > 0:
                cur_view_params_array = view_params_array[view_indices]

            batch_size = view_batch_size

            def back_project_view_batch_helper(local_view_batch, local_view_params_batch, local_pixel_indices):
                return sparse_back_project_view_batch(local_view_batch, local_view_params_batch, local_pixel_indices,
                                                      coeff_power, pixels_per_batch=pixel_batch_size)

            data_to_batch = (sinogram, cur_view_params_array)  # Apply ensure_tuple
            summed_output = sum_function_in_batches(back_project_view_batch_helper, data_to_batch, batch_size,
                                                    pixel_indices)
            return summed_output

        @partial(jax.jit, static_argnames=['coeff_power', 'pixels_per_batch'])
        def sparse_back_project_view_batch(view_batch, view_params_batch, pixel_indices, coeff_power=1,
                                           pixels_per_batch=None):
            """
            This function creates batches of pixels and collects the results to form the full sinogram.
            Also, since the geometry-specific projectors map between a batch of voxel cylinders and a single view,
            we need to map over views and add the results to get the correct back projection for each voxel.

            Args:
                view_batch (ndarray or jax array): 3D array of shape (cur_num_views, num_det_rows, num_det_cols)
                view_params_batch (jax array): 1D or 2D array of parameters, view_params_batch[i] describes view_batch[i]
                pixel_indices (ndarray or jax array): 1D array of indices into a flattened array of shape
                (num_recon_rows, num_recon_cols)
                coeff_power (int): backproject using the coefficients of (A_ij ** coeff_power).
                    Normally 1, but should be 2 when computing Hessian diagonal.

            Returns:
                jax array of size (pixel_indices.shape[0], num_recon_slices)
            """
            def back_project_pixel_batch(pixel_indices_batch):
                """
                Apply back_project_one_view_to_pixel_batch to each pixel batch and each view
                Add over the views and concatenate over the pixels.

                Args:
                    pixel_indices_batch:

                Returns:
                    jax array of size (pixel_indices.shape[0], num_recon_slices)
                """
                #
                bp_vmap = jax.vmap(back_project_one_view_to_pixel_batch, in_axes=(0, None, 0, None, None))
                per_view_voxel_values_batch = bp_vmap(view_batch, pixel_indices_batch, view_params_batch,
                                                      projector_params, coeff_power)

                voxel_values_batch = jnp.sum(per_view_voxel_values_batch, axis=0)
                return voxel_values_batch

            new_voxel_values = concatenate_function_in_batches(back_project_pixel_batch, pixel_indices, pixels_per_batch)
            return new_voxel_values

        # Set the compiled projectors
        projector_functions = (jax.jit(sparse_forward_project_fcn),
                               jax.jit(sparse_back_project_fcn, static_argnames='coeff_power'))
        self.sparse_forward_project, self.sparse_back_project = projector_functions
        self.forward_project_pixel_batch = sparse_forward_project_pixel_batch
        self.back_project_view_batch = sparse_back_project_view_batch


def concatenate_function_in_batches(function, data_to_batch, batch_size):
    """
    Apply a given function to a set of data, batching over the first index, concatenating the results along axis=0.
    The function should operate on subsets of the form data_to_batch[start:start+batch_size] when function takes a
    single input or on analogous subsets of each element data_to_batch[j] when function takes multiple
    inputs.  The output of function should be an array or tuple of arrays.  These output are concatenated along
    the leading axis.

    The shape of each array is determined by the output(s) of function, which
    is concatenated along the leading axis.  If function returns a fixed shape output, then the result has
    size given by the total number batches (full or partial) times the length of the leading axis of the output.

    Args:
        function (callable): A function to be mapped over batches of the input data.  This will be called on
        the unpacked elements of data_to_batch (after batching) using a call of the form
        output_data = function(batch) when data_to_batch is a single array or
        output_data = function(*batch) when data_to_batch is a tuple.
        data_to_batch (jax array or tuple of arrays): An array of data to be batched and sent to function. If a tuple,
        then each element should have the same size leading axis.
        batch_size (int): The maximum number of entries to process at one time.

    Returns:
        An array or tuple of arrays.
    """
    data_to_batch = ensure_tuple(data_to_batch)

    # Apply the batch projector directly to an initial batch
    num_input_points = data_to_batch[0].shape[0]
    batch_size = num_input_points if batch_size is None else batch_size
    num_remaining = num_input_points % batch_size

    # If the input is a multiple of batch_size, then we'll do a full batch, otherwise just the excess.
    initial_batch_size = batch_size if num_remaining == 0 else num_remaining

    initial_batch = [data[:initial_batch_size] for data in data_to_batch]
    output_data = function(*initial_batch)

    # Then deal with the batches if there are any
    if batch_size < num_input_points:
        def wrapped_function(arg_list):
            return function(*arg_list)

        num_batches = (num_input_points - initial_batch_size) // batch_size
        output_shape = (num_batches, batch_size,)
        data_batched = [jnp.reshape(data[initial_batch_size:], output_shape + data.shape[1:])
                        for data in data_to_batch]

        # Apply the function in batches
        output_data_batched = jax.lax.map(wrapped_function, data_batched)

        # The output data may be a single array or a tuple or list of arrays
        # First unbatch the data by reshaping the first 2 dims to be the number of points in all the batches.
        # Using tree_map, this can be done on either a single array or a tuple of arrays and get either a
        # single array or a tuple back
        output_unbatched = jax.tree_util.tree_map(unbatch, output_data_batched)

        # Now stack the first partial batch with this unbatched result
        output_data = jax.tree_util.tree_map(concatenate_arrays, *(output_data, output_unbatched))

    return output_data


def unbatch(array):
    """
    Reshape a jax array from (n0, n1, ...) to (n0*n1, ...)

    Args:
        array (jax array): array to be reshaped

    Returns:
        jax array
    """
    return jax.numpy.reshape(array, (array.shape[0] * array.shape[1],) + array.shape[2:])


def concatenate_arrays(*arrays):
    """
    Helper function to concatenate a list or tuple of arrays along the leading axis.

    Args:
        *arrays: list of arrays, with compatible dimensions arrays[j].shape[1:]

    Returns:
        array of shape (n, ) + arrays[0].shape[1:], where n is the sum over j of arrays[j].shape[0]
    """
    return jax.numpy.concatenate(arrays, axis=0)


def sum_function_in_batches(function_to_sum, data_to_batch, batch_size, extra_args=()):
    """
    Apply a given function to a set of data, batching over the first index, summing the results.
    The function should operate on subsets of the form data_to_batch[start:start+batch_size] when function takes a
    single input or on analogous subsets of each element data_to_batch[j] when function takes multiple
    inputs.  The output of function should be a scalar or fixed size array.

    Args:
        function_to_sum (callable): A function to be mapped over batches of the input data.  This will be called on
        the unpacked elements of data_to_batch (after batching) and extra_args using a call of the form
        summed_data += function_to_sum(batched_data, *fixed_data) when data_to_batch is a single array or
        summed_data += function_to_sum(*batched_data, *fixed_data) when data_to_batch is a tuple.
        data_to_batch (jax array or tuple of arrays): The data to be processed in batches.  If a tuple, then each element
        should have the same size leading axis.
        batch_size (int): The maximum batch size.
        extra_args (tuple): Any additional arguments needed by function_to_sum

    Returns:
        jax array or scalar output of function_to_sum, summed over all the elements in data_to_batch.
    """
    data_to_batch = ensure_tuple(data_to_batch)
    extra_args = ensure_tuple(extra_args)

    def add_one_batch(summed_and_fixed_data, batched_data):
        """
        Apply the externally defined function function_to_sum to the data in the tuple batched_data
        and add the result to an existing result.  The existing result is the first element in the tuple
        summed_and_fixed_data.  Any remaining elements of summed_and_fixed_data are for additional arguments
        to function_to_sum.  batched_data and fixed_data are unpacked before calling function_to sum. The
         primary functionality is summed_data += function_to_sum(*batched_data, *fixed_data)

        Args:
            summed_and_fixed_data (tuple or list): The first element is an array of the shape returned by
            function_to_sum.  This shape should not depend on batched_data.  The remaining elements are
            extra arguments to be sent to function_to_sum.
            batched_data (tuple or list):  The data for use in function_to_sum.

        Returns:
            tuple of ([summed_data, *fixed_data], None)
        """
        summed_data = summed_and_fixed_data[0]
        fixed_data = summed_and_fixed_data[1:]
        output_to_add = function_to_sum(*batched_data, *fixed_data)
        summed_data = jax.tree_util.tree_map(jnp.add, *(summed_data, output_to_add))

        return [summed_data, *fixed_data], None

    # Apply the batch projector directly to an initial batch to get the initial output
    num_input_points = data_to_batch[0].shape[0]
    batch_size = num_input_points if batch_size is None else batch_size
    num_remaining = num_input_points % batch_size
    # If the input is a multiple of batch_size, then we'll do a full batch, otherwise just the excess.
    initial_batch_size = batch_size if num_remaining == 0 else num_remaining

    initial_batch = [data[:initial_batch_size] for data in data_to_batch]
    summed_output = function_to_sum(*initial_batch, *extra_args)

    # Then deal with the batches if there are any
    if batch_size < num_input_points:
        num_batches = (num_input_points - initial_batch_size) // batch_size
        output_shape = (num_batches, batch_size,)
        data_batched = [jnp.reshape(data[initial_batch_size:], output_shape + data.shape[1:])
                        for data in data_to_batch]

        # Set up a scan over the batches.
        initial_carry = [summed_output, *extra_args]
        final_carry, _ = jax.lax.scan(add_one_batch, initial_carry, data_batched)

        summed_output = final_carry[0]

    return summed_output


def ensure_tuple(var_args):
    """
    Convert a singleton to a one-element tuple if needed, and convert a list to a tuple
    Args:
        var_args: singleton or list or tuple

    Returns:
        tuple
    """
    # Check if var_args is already a tuple
    if isinstance(var_args, tuple):
        return var_args
    # Check if var_args is a list
    elif isinstance(var_args, list):
        return tuple(var_args)
    # Assume var_args is a single item if it's neither a list nor a tuple
    else:
        return (var_args, )


import mbirjax as mj
import numpy as np

if __name__ == "__main__":

    num_views = 64
    num_det_rows = 40
    num_det_channels = 128

    start_angle = -np.pi
    end_angle = np.pi

    # Initialize sinogram
    sinogram_shape = (num_views, num_det_rows, num_det_channels)
    angles = np.linspace(start_angle, end_angle, num_views, endpoint=False)
    ct_model_for_generation = mj.ParallelBeamModel(sinogram_shape, angles)

    phantom_shape = ct_model_for_generation.get_params('recon_shape')
    phantom = mj.generate_3d_shepp_logan_low_dynamic_range(phantom_shape)
    # mj.slice_viewer(phantom, phantom + 0.1 * np.random.rand(*phantom.shape) - 0.05, phantom + 0.05 * np.random.randn(*phantom.shape), slice_axis=(0, 0, 0))

    sinogram = ct_model_for_generation.forward_project(phantom)

    recon_fbp = ct_model_for_generation.direct_recon(sinogram)
    recon, recon_dict = ct_model_for_generation.recon(sinogram)

    mj.slice_viewer(recon, sinogram, data_dicts=[recon_dict, None], slice_axis=[2, 0])

    ct_model_for_generation.save_recon_hdf5(filepath='./test_fbp.h5', recon=recon_fbp)
    ct_model_for_generation.save_recon_hdf5(filepath='./test_mbir.h5', recon=recon, recon_dict=recon_dict)

    # Test viewer

    mj.slice_viewer(None, None, vmin=0.0, vmax=1.0, slice_axis=(2, 2), title='Right-click to load images')



import numpy as np
import jax.numpy as jnp
import mbirjax as mj
import unittest


class TestFBPReconstruction(unittest.TestCase):
    """
    Unit test for FBP reconstruction.
    Specifically tests FBP for parallel beam geometry and cone-beam geometry (FDK method) with a planar detector panel.
    Tests the accuracy of the reconstruction against the 3 metrics: NRMSE, max_diff, and pct_95.
    """

    def setUp(self):
        """Set up parameters and initialize models before each test."""
        # Sinogram parameters
        self.num_views = 128
        self.num_det_rows = 128
        self.num_det_channels = 128
        self.sinogram_shape = (self.num_views, self.num_det_rows, self.num_det_channels)

        # Geometry parameters - FBP
        self.fbp_angles =  jnp.linspace(-jnp.pi * (1/2), jnp.pi * (1/2), self.num_views, endpoint=False)

        # Geometry parameters - FDK
        self.source_detector_dist = 4 * self.num_det_channels
        self.source_iso_dist = self.source_detector_dist / 2
        start_angle = -jnp.pi  # For testing purposes, we use a full 360 degrees
        end_angle = jnp.pi
        self.fdk_angles = jnp.linspace(start_angle, end_angle, self.num_views, endpoint=False)

        # Initialize both models
        self.parallel_model = mj.ParallelBeamModel(self.sinogram_shape, self.fbp_angles)
        self.cone_model = mj.ConeBeamModel(self.sinogram_shape,
                                                self.fdk_angles,
                                                source_detector_dist=self.source_detector_dist,
                                                source_iso_dist=self.source_iso_dist)

        # Generate 3D Shepp-Logan phantom and sinogram
        phantom_shape = self.parallel_model.get_params('recon_shape')
        self.fbp_phantom = mj.generate_3d_shepp_logan_low_dynamic_range(phantom_shape)
        self.fbp_sino = self.parallel_model.forward_project(self.fbp_phantom)

        phantom_shape = self.cone_model.get_params('recon_shape')
        self.fdk_phantom = mj.generate_3d_shepp_logan_low_dynamic_range(phantom_shape)
        self.fdk_sino = self.cone_model.forward_project(self.fdk_phantom)

        # Set tolerances for the metrics - FBP is deterministic, so results should never go above these.
        self.fbp_tolerances = {'nrmse': 0.21, 'max_diff': 0.41, 'pct_95': 0.064}
        self.fdk_tolerances = {'nrmse': 0.27, 'max_diff': 0.51, 'pct_95': 0.1}

    def test_fbp_reconstruction(self):
        """Test the FBP reconstruction against the defined tolerances."""
        # Perform FBP reconstruction
        filter_name = "ramp"
        recon = self.parallel_model.direct_recon(self.fbp_sino, filter_name=filter_name)

        # Compute the statistics
        max_diff = np.amax(np.abs(self.fbp_phantom - recon))
        nrmse = np.linalg.norm(recon - self.fbp_phantom) / np.linalg.norm(self.fbp_phantom)
        pct_95 = np.percentile(np.abs(recon - self.fbp_phantom), 95)

        # Verify that the computed stats are within tolerances
        self.assertTrue(max_diff < self.fbp_tolerances['max_diff'], f"Max difference too high: {max_diff}")
        self.assertTrue(nrmse < self.fbp_tolerances['nrmse'], f"NRMSE too high: {nrmse}")
        self.assertTrue(pct_95 < self.fbp_tolerances['pct_95'], f"95th percentile difference too high: {pct_95}")

    def test_fdk_reconstruction(self):
        """Test the FDK reconstruction against the defined tolerances."""
        # Perform FBP reconstruction
        filter_name = "ramp"
        recon = self.cone_model.direct_recon(self.fdk_sino, filter_name=filter_name)

        # Compute the statistics
        max_diff = np.amax(np.abs(self.fdk_phantom - recon))
        nrmse = np.linalg.norm(recon - self.fdk_phantom) / np.linalg.norm(self.fdk_phantom)
        pct_95 = np.percentile(np.abs(recon - self.fdk_phantom), 95)

        # Verify that the computed stats are within tolerances
        self.assertTrue(max_diff < self.fdk_tolerances['max_diff'], f"Max difference too high: {max_diff}")
        self.assertTrue(nrmse < self.fdk_tolerances['nrmse'], f"NRMSE too high: {nrmse}")
        self.assertTrue(pct_95 < self.fdk_tolerances['pct_95'], f"95th percentile difference too high: {pct_95}")

if __name__ == '__main__':
    unittest.main()


import jax
import os
import psutil


def get_memory_stats(print_results=True, file=None):
    # Get all GPU devices
    gpus = [device for device in jax.devices() if 'cpu' not in device.device_kind.lower()]

    memory_stats_per_processor = []

    # Collect memory info for gpus
    for gpu in gpus:
        # Memory info returns total_memory and available_memory in bytes
        gpu_stats = gpu.memory_stats()
        memory_stats = dict()
        memory_stats['id'] = 'GPU ' + str(gpu.id)
        memory_stats['bytes_in_use'] = gpu_stats['bytes_in_use']
        memory_stats['peak_bytes_in_use'] = gpu_stats['peak_bytes_in_use']
        memory_stats['bytes_limit'] = gpu_stats['bytes_limit']

        memory_stats_per_processor.append(memory_stats)

    # Then add info for the CPU
    memory_stats = dict()
    # Get the current process ID
    pid = os.getpid()

    # Create a process object using the current PID
    current_process = psutil.Process(pid)

    # Get the memory usage
    memory_info = current_process.memory_full_info()
    memory_stats['id'] = 'CPU'

    # memory_info.rss will return the Resident Set Size (RSS)
    # which is the non-swapped physical memory the process has used
    memory_stats['peak_bytes_in_use'] = memory_info.rss
    memory_stats['bytes_in_use'] = memory_info.uss  # This is the 'Unique Set Size' used by the process
    # Get the virtual memory statistics
    mem = psutil.virtual_memory()

    # Available physical memory (excluding swap)
    memory_stats['bytes_limit'] = mem.available
    memory_stats_per_processor.append(memory_stats)

    if print_results:
        for memory_stats in memory_stats_per_processor:
            print(memory_stats['id'], file=file)
            for tag in ['bytes_in_use', 'peak_bytes_in_use', 'bytes_limit']:
                cur_value = memory_stats[tag] / (1024 ** 3)
                extra_space = ' ' * (21 - len(tag) - len(str(int(cur_value))))
                print(f'  {tag}:{extra_space}{cur_value:.3f}GB', file=file)

    return memory_stats_per_processor



# -*- coding: utf-8 -*-
"""demo_4_wrong_rotation_direction.ipynb

Automatically generated by Colab.

Original file is located at
    https://colab.research.google.com/drive/1Gd-fMm3XK1WBsuJUklHdZ-4jjsvdpeIT

**MBIRJAX: Wrong Rotation Direction Demo**

See the [MBIRJAX documentation](https://mbirjax.readthedocs.io/en/latest/) for an overview and details.  

This script demonstrates the effects of an incorrect rotation direction when using cone beam projection.  

For this demo, we do a baseline reconstruction with the correct specification of angles, then reverse the order of the angles and perform another reconstruction.  

Since this problem doesn't arise with parallel beam because of symmetry, we show this only for cone beam.   

See [demo_1_shepp_logan.py](https://colab.research.google.com/drive/1zG_H6CDjuQxeMRQHan3XEyX2YVKcSSNC) for the basic steps of synthetic sinogram generation and reconstruction.

Select a GPU as runtime type for best performance.
"""

# Commented out IPython magic to ensure Python compatibility.
# %pip install mbirjax

import numpy as np
import pprint
import time
import jax.numpy as jnp
import mbirjax as mj

"""**Set the geometry parameters**"""

# Set parameters for the problem size.
num_views = 64
num_det_rows = 40
num_det_channels = 128

sharpness = 0.0

# For cone beam geometry, we need to describe the distances source to detector and source to rotation axis.
# np.Inf is an allowable value, in which case this is essentially parallel beam
source_detector_dist = 4 * num_det_channels
source_iso_dist = source_detector_dist

# For cone beam reconstruction, we need a little more than 180 degrees for full coverage.
detector_cone_angle = 2 * np.arctan2(num_det_channels / 2, source_detector_dist)
start_angle = -(np.pi + detector_cone_angle) * (1/2)
end_angle = (np.pi + detector_cone_angle) * (1/2)

"""**Data generation:** For demo purposes, we create a phantom and then project it to create a sinogram.

For this demo we use the default recon shape for cone beam.

Note:  the sliders on the viewer won't work in notebook form.  For that you'll need to run the python code with an interactive matplotlib backend, typcially using the command line or a development environment like Spyder or Pycharm to invoke python.  

"""

# Choose the geometry type - for wrong rotation, only conebeam geometry will have a problem.
model_type = 'cone'
object_type = 'shepp-logan'  # 'shepp-logan' or 'cube'


# Generate simulated data
# In a real application you would not have the phantom, but we include it here for later display purposes
phantom, sinogram, params = mj.generate_demo_data(object_type=object_type, model_type=model_type,
                                                  num_views=num_views, num_det_rows=num_det_rows,
                                                  num_det_channels=num_det_channels)
angles = params['angles']


# View sinogram
title='Original sinogram'
mj.slice_viewer(sinogram, title=title, slice_axis=0, slice_label='View')

"""**Do a baseline reconstruction**

Here we do a standard reconstruction with the correct specification of angles.  
"""

print('\nStarting recon with correct rotation\n')
ct_model = mj.ConeBeamModel(sinogram.shape, angles, source_detector_dist=source_detector_dist,
                                            source_iso_dist=source_iso_dist)
ct_model.set_params(sharpness=sharpness)

recon_correct, recon_dict_correct = ct_model.recon(sinogram)

recon_params_correct = recon_dict_correct['recon_params']
pprint.pprint(recon_params_correct, compact=True)

"""**Do a reconstruction with angles in reverse order.**"""

print('\nStarting recon with incorrect rotation\n')
angles_reversed = angles[::-1]
ct_model = mj.ConeBeamModel(sinogram.shape, angles_reversed, source_detector_dist=source_detector_dist,
                                            source_iso_dist=source_iso_dist)
recon_incorrect, recon_dict_incorrect = ct_model.recon(sinogram)

recon_params_incorrect = recon_dict_incorrect['recon_params']
pprint.pprint(recon_params_incorrect, compact=True)

"""**Display the results with correct and incorrect angles.**"""

title = 'Correct rotation recon (left) vs incorrect rotation recon (right)'
title += '\nThe incorrect angle specification leads to shape distortion and top/bottom reflection.'
mj.slice_viewer(recon_correct, recon_incorrect, data_dicts=[recon_dict_correct, recon_dict_incorrect], title=title, vmin=0.0, vmax=1.0)

"""**Next:** Try changing some of the parameters and re-running or try [some of the other demos](https://mbirjax.readthedocs.io/en/latest/demos_and_faqs.html).  """


import os
import numpy as np
import jax.numpy as jnp
import jax
import time
from functools import partial

os.environ['XLA_PYTHON_CLIENT_MEM_FRACTION'] = '0.98'

def set_sinogram_parameters():
    # Specify sinogram info
    num_views = 600
    num_det_rows = 1500
    num_det_channels = 2000
    return num_views, num_det_rows, num_det_channels

def sparse_forward_project(voxel_values, indices, sinogram_shape, recon_shape, angles, output_device, worker):
    """
    Batch the views (angles) and voxels/indices, send batches to the GPU to project, and collect the results.
    """
    max_views_per_batch = 200
    max_pixels_per_batch = 8000
    num_pixels_to_exclude = 1000

    indices = indices[:len(indices)-num_pixels_to_exclude]
    angles = jax.device_put(angles, device=worker)

    # Batch the views and pixels
    num_views = len(angles)
    view_batch_indices = jnp.arange(num_views, step=max_views_per_batch)
    view_batch_indices = jnp.concatenate([view_batch_indices, num_views * jnp.ones(1, dtype=int)])

    num_pixels = len(indices)
    pixel_batch_indices = jnp.arange(num_pixels, step=max_pixels_per_batch)
    pixel_batch_indices = jnp.concatenate([pixel_batch_indices, num_pixels * jnp.ones(1, dtype=int)])

    # Create the output sinogram
    sinogram = []

    # Loop over the view batches
    for j, view_index_start in enumerate(view_batch_indices[:-1]):
        # Send a batch of views to worker
        view_index_end = view_batch_indices[j+1]
        cur_view_batch = jnp.zeros([view_index_end-view_index_start, sinogram_shape[1], sinogram_shape[2]],
                                   device=worker)
        cur_view_params_batch = angles[view_index_start:view_index_end]
        if j == 0:
            get_memory_stats()
        print('Starting view block {} of {}.'.format(j+1, view_batch_indices.shape[0]-1))

        # Loop over pixel batches
        for k, pixel_index_start in enumerate(pixel_batch_indices[:-1]):
            # Send a batch of pixels to worker
            pixel_index_end = pixel_batch_indices[k+1]
            cur_voxel_batch, cur_index_batch = jax.device_put([voxel_values[pixel_index_start:pixel_index_end],
                                                              indices[pixel_index_start:pixel_index_end]],
                                                              worker)
            if len(cur_index_batch) < max_pixels_per_batch:
                cur_voxel_batch = jnp.concatenate([cur_voxel_batch, jnp.zeros([max_pixels_per_batch-len(cur_index_batch), sinogram_shape[1],], device=worker)])
                cur_index_batch = jnp.concatenate([cur_index_batch, jnp.zeros([max_pixels_per_batch-len(cur_index_batch)], device=worker)])

            def forward_project_pixel_batch_local(view, angle):
                # Add the forward projection to the given existing view
                return forward_project_pixel_batch_to_one_view(cur_voxel_batch, cur_index_batch, angle, view,
                                                               sinogram_shape, recon_shape)

            view_map = jax.vmap(forward_project_pixel_batch_local)
            # print(jax.make_jaxpr(view_map)(cur_view_batch, cur_view_params_batch))
            # input('Enter to continue')
            # a = jax.jit(view_map).lower(cur_view_batch, cur_view_params_batch).compiler_ir('hlo')
            # with open("outfile.dot", "w") as f:
            #     f.write(a.as_hlo_dot_graph())
            # dot outfile.dot  -Tpng > outfile.png
            # or
            # dot -Tps outfile.dot -o outfile.ps
            # ps2pdf outfile.ps
            # print(jax.jit(view_map).lower(cur_view_batch, cur_view_params_batch).compile().as_text())
            cur_view_batch = view_map(cur_view_batch, cur_view_params_batch)

        sinogram.append(jax.device_put(cur_view_batch, output_device))
    sinogram = jnp.concatenate(sinogram)
    return sinogram


@partial(jax.jit, static_argnames=['sinogram_shape', 'recon_shape'], donate_argnames='sinogram_view')
def forward_project_pixel_batch_to_one_view(voxel_values, pixel_indices, angle, sinogram_view,
                                            sinogram_shape, recon_shape):
    """
    Apply a parallel beam transformation to a set of voxel cylinders. These cylinders are assumed to have
    slices aligned with detector rows, so that a parallel beam maps a cylinder slice to a detector row.
    This function returns the resulting sinogram view.

    """
    # Get all the geometry parameters - we use gp since geometry parameters is a named tuple and we'll access
    # elements using, for example, gp.delta_det_channel, so a longer name would be clumsy.
    num_views, num_det_rows, num_det_channels = sinogram_shape
    psf_radius = 1
    delta_voxel = 1

    # Get the data needed for horizontal projection
    n_p, n_p_center, W_p_c, cos_alpha_p_xy = compute_proj_data(pixel_indices, angle, sinogram_shape, recon_shape)
    L_max = jnp.minimum(1, W_p_c)

    # Do the projection
    for n_offset in jnp.arange(start=-psf_radius, stop=psf_radius+1):
        n = n_p_center + n_offset
        abs_delta_p_c_n = jnp.abs(n_p - n)
        L_p_c_n = jnp.clip((W_p_c + 1) / 2 - abs_delta_p_c_n, 0, L_max)
        A_chan_n = delta_voxel * L_p_c_n / cos_alpha_p_xy
        A_chan_n *= (n >= 0) * (n < num_det_channels)
        sinogram_view = sinogram_view.at[:, n].add(A_chan_n.reshape((1, -1)) * voxel_values.T)

    return sinogram_view

def compute_proj_data(pixel_indices, angle, sinogram_shape, recon_shape):
    """
    Compute the quantities n_p, n_p_center, W_p_c, cos_alpha_p_xy needed for vertical projection.
    """

    cosine = jnp.cos(angle)
    sine = jnp.sin(angle)

    delta_voxel = 1.0
    dvc = delta_voxel
    dvs = delta_voxel
    dvc *= cosine
    dvs *= sine

    num_views, num_det_rows, num_det_channels = sinogram_shape

    # Convert the index into (i,j,k) coordinates corresponding to the indices into the 3D voxel array
    row_index, col_index = jnp.unravel_index(pixel_indices, recon_shape[:2])

    y_tilde = dvs * (row_index - (recon_shape[0] - 1) / 2.0)
    x_tilde = dvc * (col_index - (recon_shape[1] - 1) / 2.0)

    x_p = x_tilde - y_tilde

    det_center_channel = (num_det_channels - 1) / 2.0  # num_of_cols

    # Calculate indices on the detector grid
    n_p = x_p + det_center_channel
    n_p_center = jnp.round(n_p).astype(int)

    # Compute cos alpha for row and columns
    cos_alpha_p_xy = jnp.maximum(jnp.abs(cosine), jnp.abs(sine))

    # Compute projected voxel width along columns and rows (in fraction of detector size)
    W_p_c = cos_alpha_p_xy

    proj_data = (n_p, n_p_center, W_p_c, cos_alpha_p_xy)

    return proj_data


def get_memory_stats(print_results=True, file=None):
    # Get all GPU devices
    gpus = [device for device in jax.devices() if 'cpu' not in device.device_kind.lower()]

    # Collect memory info for gpus
    for gpu in gpus:
        # Memory info returns total_memory and available_memory in bytes
        gpu_stats = gpu.memory_stats()
        memory_stats = dict()
        memory_stats['id'] = 'GPU ' + str(gpu.id)
        memory_stats['bytes_in_use'] = gpu_stats['bytes_in_use']
        memory_stats['peak_bytes_in_use'] = gpu_stats['peak_bytes_in_use']
        memory_stats['bytes_limit'] = gpu_stats['bytes_limit']

        print(memory_stats['id'], file=file)
        for tag in ['bytes_in_use', 'peak_bytes_in_use', 'bytes_limit']:
            cur_value = memory_stats[tag] / (1024 ** 3)
            extra_space = ' ' * (21 - len(tag) - len(str(int(cur_value))))
            print(f'  {tag}:{extra_space}{cur_value:.3f}GB', file=file)



def main():
    """
    This is a script to develop, debug, and tune the parallel beam projector
    """

    # Specify sinogram info
    num_views, num_det_rows, num_det_channels = set_sinogram_parameters()
    start_angle = 0
    end_angle = jnp.pi
    sinogram_shape = (num_views, num_det_rows, num_det_channels)
    angles = jnp.linspace(start_angle, end_angle, num_views, endpoint=False)

    output_device = jax.devices('cpu')[0]
    try:
        worker = jax.devices('gpu')[0]
        use_gpu = True
    except RuntimeError:
        worker = jax.devices('cpu')[0]
        use_gpu = False

    # Generate phantom - all zero except a small cube
    recon_shape = (num_det_channels, num_det_channels, num_det_rows)
    num_recon_rows, num_recon_cols, num_recon_slices = recon_shape[:3]
    with jax.default_device(output_device):
        phantom = jnp.zeros(recon_shape)  #mj.gen_cube_phantom(recon_shape)
        i, j, k = recon_shape[0]//3, recon_shape[1]//2, recon_shape[2]//2
        phantom = phantom.at[i:i+5, j:j+5, k:k+5].set(1.0)

        # Generate indices of pixels and sinogram data
        # Determine the 2D indices within the RoR
        max_index_val = num_recon_rows * num_recon_cols
        indices = np.arange(max_index_val, dtype=np.int32)
        indices = jnp.array(indices)
        voxel_values = phantom.reshape((-1,) + recon_shape[2:])[indices]

    print('Starting forward projection')
    voxel_values, indices = jax.device_put([voxel_values, indices], output_device)
    t0 = time.time()
    sinogram = sparse_forward_project(voxel_values, indices, sinogram_shape, recon_shape, angles,
                                      output_device=output_device, worker=worker)
    print('Elapsed time:', time.time() - t0)

    # Determine resulting number of views, slices, and channels and image size
    print('Sinogram shape: {}'.format(sinogram.shape))
    if use_gpu:
        print('Memory stats after forward projection')
        get_memory_stats(print_results=True)

    # import mbirjax
    # mj.slice_viewer(sinogram, slice_axis=0)


if __name__ == "__main__":
    main()